# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    import sys
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve()) 
    sys.path.append(os.getcwd())
    print('__main__ - current working directory:', os.getcwd())
    
    

import pandas as pd
import numpy as np
import math
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from collections import OrderedDict
import scipy.interpolate as interp
import scipy.stats as sci
from scipy.stats import norm
from scipy import interpolate
from numpy import array
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import warnings
warnings.filterwarnings('ignore')


class Conrad:
    """Vanna-Volga Implied Volatility method used to price an FX option"""

    def __init__(self, S0, strike, t, importfile):
        self.dc = 360  # day count convention
        self.S0 = S0
        self.strike = strike
        self.t = t
        self.importfile = importfile

    def foreign_rate(self, fp, Rates):
        xf = [] * len(fp)
        Rates_array = np.array(Rates).squeeze()
        Rates_interp = interp.interp1d(np.arange(Rates_array.size), Rates_array, kind='cubic')
        Rates_compress = Rates_interp(np.linspace(0, Rates_array.size - 1, fp.size))
        for i in range(len(Rates_compress)):
            xf.append(((fp[i] / self.S0) + 1) * (1 + Rates_compress[i]) - 1)

        return Rates_compress, xf

    def k25_data(self, T, sig25DeltaPut, sigATM, sig25DeltaCall, xd, xf):
        """Calculate the strikes K25DeltaPut, KATM, K25DeltaCall"""

        # Pricing Inputs
        Dd = np.exp(-np.multiply(xd, T))
        Df = np.exp(-np.multiply(xf, T))
        delta = 0.25
        alpha = -sci.norm.ppf(delta * np.reciprocal(Df))
        mu = xd - xf
        F = self.S0 * np.exp(np.multiply(mu, T))

        # Strikes
        KATM = np.array([])
        K25DeltaCall = np.array([])
        K25DeltaPut = np.array([])

        for i in range(len(T)):
            K25DeltaPut = np.append(K25DeltaPut, F[i] * math.exp(
                -(alpha[i] * sig25DeltaPut[i] * math.sqrt(T[i])) + (xd[i] - xf[i] + 0.5 * sig25DeltaPut[i] ** 2) * T[i]))
            KATM = np.append(KATM, F[i] * math.exp(xd[i] - xf[i] + 0.5 * T[i] * (sigATM[i]) ** 2))
            K25DeltaCall = np.append(K25DeltaCall, F[i] * math.exp(
                alpha[i] * sig25DeltaCall[i] * math.sqrt(T[i]) + (xd[i] - xf[i] + 0.5 * sig25DeltaCall[i] ** 2) * T[i]))

        return K25DeltaPut, KATM, K25DeltaCall, F

    def d1(self, F, K, sig, T, xd, xf):
        d1_val = (np.log(F / K) + (xd - xf + 0.5 * (sig ** 2)) * T) / (sig * np.sqrt(T))

        return d1_val

    def d2(self, F, K, sig, T, xd, xf):
        d2_val = self.d1(F, K, sig, T, xd, xf) - sig * np.sqrt(T)

        return d2_val

    def bs_price(self, F, K, T, xd, xf, sig):
        a = norm.cdf(self.d1(F, K, sig, T, xd, xf))
        b = self.d1(F, K, sig, T, xd, xf)

        bs_val = norm.cdf(self.d1(F, K, sig, T, xd, xf)) * F * math.exp(-xf * T) - norm.cdf(
            self.d2(F, K, sig, T, xd, xf)) * F * math.exp(-xd * T)

        return bs_val

    def vega(self, F, K, T, xd, xf, sig):
        np = math.exp(-self.d1(F, K, sig, T, xd, xf) ** 2 / 2) / math.sqrt(2 * math.pi)
        vega_val = (F * math.exp(-xf * T) * math.sqrt(T) * np) / 100

        return vega_val

    def delta(self, F, K, T, xd, xf, sig):
        delta_val = math.exp(xf * T) * norm.cdf(self.d1(F, K, sig, T, xd, xf))

        return delta_val

    def implied_price(self, F, K, T, K1, K2, K3, sig1, sig2, sig3, xd, xf):
        v_v1 = self.vega(F, K, T, xd, xf, sig2) / self.vega(F, K1, T, xd, xf, sig1)
        v_v2 = self.vega(F, K, T, xd, xf, sig2) / self.vega(F, K2, T, xd, xf, sig2)
        v_v3 = self.vega(F, K, T, xd, xf, sig2) / self.vega(F, K3, T, xd, xf, sig3)

        x1 = v_v1 * ((math.log(K2 / K) * math.log(K3 / K)) / (math.log(K2 / K1) * math.log(K3 / K1)))
        x2 = v_v2 * ((math.log(K / K1) * math.log(K3 / K)) / (math.log(K2 / K1) * math.log(K3 / K2)))
        x3 = v_v3 * ((math.log(K / K1) * math.log(K / K2)) / (math.log(K3 / K1) * math.log(K3 / K2)))

        a = self.bs_price(F, K2, T, xd, xf, sig2)

        vv_price = self.bs_price(F, K2, T, xd, xf, sig2) + x1 * (
        self.bs_price(F, K1, T, xd, xf, sig1) - self.bs_price(F, K1, T, xd, xf, sig2))
        + x3 * (self.bs_price(F, K3, T, xd, xf, sig3) - self.bs_price(F, K3, T, xd, xf, sig2))

        return vv_price

    def market_price(self, F, K, T, K1, K2, K3, sig1, sig2, sig3, xd, xf, j):
        sig = self.implied_vol(F, K, T, K1, K2, K3, sig1, sig2, sig3, xd, xf, j)
        market_val = self.bs_price(F, K, T, xd, xf, sig)

        return market_val

    def implied_vol(self, F, K, T, K1, K2, K3, sig1, sig2, sig3, xd, xf, j):
        x1 = (math.log(K2 / K) * math.log(K3 / K)) / (math.log(K2 / K1) * math.log(K3 / K1))
        x2 = (math.log(K / K1) * math.log(K3 / K)) / (math.log(K2 / K1) * math.log(K3 / K2))
        x3 = (math.log(K / K1) * math.log(K / K2)) / (math.log(K3 / K1) * math.log(K3 / K2))

        D1 = (x1 * sig1 + x2 * sig2 + x3 * sig3) - sig2

        D2 = (x1 * self.d1(F, K1, sig1, T, xd, xf) * self.d2(F, K1, sig1, T, xd, xf) * (sig1 - sig2) ** 2) \
             + (x2 * self.d1(F, K2, sig2, T, xd, xf) * self.d2(F, K2, sig2, T, xd, xf) * (sig2 - sig2) ** 2) \
             + (x3 * self.d1(F, K3, sig3, T, xd, xf) * self.d2(F, K3, sig3, T, xd, xf) * (sig3 - sig2) ** 2)

        sig = sig2 + (-sig2 + np.sqrt(
            sig2 ** 2 + self.d1(F, K, sig2, T, xd, xf) * self.d2(F, K, sig2, T, xd, xf) * (2 * sig2 * D1 + D2))) \
                     / (self.d1(F, K, sig2, T, xd, xf) * self.d2(F, K, sig2, T, xd, xf))

        return sig

    def interp_vv(self, row_find, col_find, flat_row_array, flat_col_array, z_matrix):
        GD = interpolate.griddata((flat_row_array, flat_col_array), z_matrix,
                                  ([row_find], [col_find]), method='cubic')

        return GD

    def create_plots(self, F, T, K25DeltaPut, KATM, K25DeltaCall, sig25DeltaPut, sigATM, sig25DeltaCall, xd,
                     xf):
        m = len(T[:-1])
        K_array = np.linspace(0.8 * self.S0, 1.2 * self.S0, m)


        implied_matrix = np.zeros((m, m), dtype=float)
        market_price_matrix = np.zeros((m, m), dtype=float)
        price_matrix = np.zeros((m, m), dtype=float)
        delta_matrix = np.zeros((m, m), dtype=float)

        for i in range(K_array.size):
            for j in range(m):
                implied_matrix[i][j] = self.implied_vol(F[j], K_array[i], T[j], K25DeltaPut[j], KATM[j],
                                                        K25DeltaCall[j],
                                                        sig25DeltaPut[j], sigATM[j], sig25DeltaCall[j], xd[j], xf[j], j)

        implied_matrix = pd.DataFrame(implied_matrix)
        implied_matrix.fillna(method='ffill', axis=1, inplace=True)

        for i in range(K_array.size):
            for j in range(m):
                price_matrix[i][j] = self.implied_price(F[j], K_array[i], T[j], K25DeltaPut[j], KATM[j],
                                                        K25DeltaCall[j],
                                                        sig25DeltaPut[j], sigATM[j], sig25DeltaCall[j], xd[j], xf[j])
                market_price_matrix[i][j] = self.market_price(F[j], K_array[i], T[j], K25DeltaPut[j], KATM[j],
                                                              K25DeltaCall[j],
                                                              sig25DeltaPut[j], sigATM[j], sig25DeltaCall[j], xd[j],
                                                              xf[j], j)
                delta_matrix[i][j] = self.delta(F[j], K_array[i], T[j], xd[j], xf[j], implied_matrix[i][j])

        #  Fill NAs
        implied_matrix = pd.DataFrame(implied_matrix)
        price_matrix = pd.DataFrame(price_matrix)
        delta_matrix = pd.DataFrame(delta_matrix)
        implied_matrix.fillna(method='ffill', axis=1, inplace=True)
        price_matrix.fillna(method='ffill', axis=1, inplace=True)
        delta_matrix.fillna(method='ffill', axis=1, inplace=True)

        #  Flatten the matrices for use with griddata
        implied_flat = implied_matrix.to_numpy().ravel()
        price_flat = price_matrix.to_numpy().ravel()
        delta_flat = delta_matrix.to_numpy().ravel()

        KK, TT = np.meshgrid(np.array(K_array), np.array(T[:-1]), indexing='ij')

        option_vol = interpolate.griddata((KK.ravel(), TT.ravel()), implied_flat,
                                  ([self.t], [self.strike]), method='nearest')

        option_premium = interpolate.griddata((KK.ravel(), TT.ravel()), price_flat,
                                  ([self.t], [self.strike]), method='nearest')

        option_delta = interpolate.griddata((KK.ravel(), TT.ravel()), delta_flat,
                                  ([self.t], [self.strike]), method='nearest')

        #  Gridspec
        gs = gridspec.GridSpec(2, 2)

        #  3D Meshgrid of T, K and Volatility
        X = T[:-1]
        Y = K_array
        X, Y = np.meshgrid(X, Y)
        Z = 100*implied_matrix
        fig = plt.figure()
        ax1 = fig.add_subplot(gs[:-1, :], projection='3d')
        ax1.set_xlabel('Tenor (yrs)')
        ax1.set_ylabel('Strike')
        ax1.set_zlabel('Implied Volatility (%)')
        plt.suptitle('Option Greeks')
        surf = ax1.plot_surface(X, Y, Z, cmap=cm.gnuplot, linewidth=0)

        #  Implied Volatility vs Strikes at 3M
        X = K_array
        Y = 100*implied_matrix[:][7]
        ax2 = fig.add_subplot(gs[-1, 0])
        ax2.set_xlabel('Strike')
        ax2.set_ylabel('Implied Volatility (%)')
        plt.title('Implied Volatility vs Strikes at 3M')
        f = ax2.plot(X, Y)

        #  Option Premium vs Strikes at 3M
        X = K_array
        Y = price_matrix[:][7]
        ax3 = fig.add_subplot(gs[-1, -1])
        ax3.set_xlabel('Strike')
        ax3.set_ylabel('C(K) (USD)')
        plt.title('Option Premium vs Strikes at 3M')
        g = ax3.plot(X, Y, 'r')
        fig.set_size_inches(13, 10)
        plt.show()

        return 'Option Volatility: {}%, Option Premium: {} USD, Option delta: {}'.format(option_vol*100, option_premium, option_delta)

    def FXOptionPrice(self):

        # Import Excel file
        xlsx_file = pd.ExcelFile(self.importfile)

        # Volatility Data
        vol_data = xlsx_file.parse('Volatility')

        # US Yield Data
        swaps_futures = xlsx_file.parse('US_Yield')
        Rates = np.array(swaps_futures['Value']) / 100
        T_Rates = pd.DataFrame({'X': swaps_futures['Term']})

        # Forward Points data
        fp_data = xlsx_file.parse('Forward_Points')
        fp = np.array(fp_data['Value']) / 10000
        T_fp = pd.DataFrame({'X': fp_data['Term']})

        T = pd.DataFrame({'X': vol_data['Term'].unique()})  # terms

        # dictionary for excel tabulated terms converted to day/360 format
        super_dict = [('SWAP', ''), ('WK', '*7'), ('MO', '*30'), ('M', '*30'), ('D', '*1'), ('DY', '*1'),
                      ('W', '*7'), ('YR', '*360'), ('Y', '*360'), (' ', '')]
        super_dict = OrderedDict(super_dict)

        # Modify the terms to daycount and store the vols in numpy arrays
        T = T.replace(super_dict, regex=True)
        T_Rates = T_Rates.replace(super_dict, regex=True)
        T_fp = T_fp.replace(super_dict, regex=True)

        T = T.X.apply(lambda x: eval(str(x))) / self.dc
        T_Rates = T_Rates.X.apply(lambda x: eval(str(x))) / self.dc
        T_fp = T_fp.X.apply(lambda x: eval(str(x))) / self.dc

        sigATM = np.array(vol_data[vol_data['Strike'] == 'ATM']['Value']) / 100  # ATM volatilities
        sig25DeltaRR = np.array(
            vol_data[vol_data['Strike'] == '25 Delta Risk Reversal']['Value']) / 100  # ATM volatilities
        sig25DeltaBF = np.array(vol_data[vol_data['Strike'] == '25 Delta Butterfly']['Value']) / 100  # ATM volatilities
        sig25DeltaCall = sig25DeltaBF + sigATM + sig25DeltaRR / 2
        sig25DeltaPut = sig25DeltaBF + sigATM - sig25DeltaRR / 2

        xd, xf = self.foreign_rate(fp, Rates)
        K25DeltaPut, KATM, K25DeltaCall, F = self.k25_data(T, sig25DeltaPut, sigATM, sig25DeltaCall, xd, xf)
        option_data = self.create_plots(F, T, K25DeltaPut, KATM, K25DeltaCall, sig25DeltaPut, sigATM, sig25DeltaCall, xd, xf)

        return option_data

def main():
    # Example Input
    # FXOptionPrice(1.1609, 1.2, 30 / 360, 'Case Study 2017 - Pricing Engine.xlsx')

    validus_answer = Conrad(1.1609, 1.1, 5, os.getcwd() + '\\pricing_engine\\in_progress\\Case Study 2017 - Pricing Engine.xlsx').FXOptionPrice()
    print(validus_answer)
    plt.show()

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-


import os
import pathlib

if __name__ == "__main__":
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve()) # path to ./frm/ 
    print(__file__.split('\\')[-1], os.getcwd()) 

import numpy as np
import pandas as pd    
from dataclasses import dataclass, field, InitVar 
from typing import Optional
import datetime as dt
import calendar

from market_data.iban_ccys import VALID_CCYS
from schedule.business_day_calendar import get_calendar
from schedule.daycounter import DayCounter, VALID_DAY_COUNT_BASIS
        
#%%    

#import warnings
#warnings.formatwarning = custom_formatwarning


def isIMMDate(date: dt.date):
           # Returns boolean on if date is an IMM date
           # IMM (International Monetary Market) dates are the third Wednesday of March, June, September and December 
           # (i.e., between the 15th and 21st, whichever such day is a Wednesday).
           if date.month in {3,6,9,12} and 15 <= date.day <= 21 and date.weekday() == 2:
               return True
           else:
               return False


def getIMMDate(year: int, month: int):
    # Get date of the 3rd Wednesday in the specified month and year 
    c = calendar.Calendar()
    monthcal = c.monthdatescalendar(year,month)
    
    imm_date = [day for week in monthcal for day in week if \
                     day.weekday() == calendar.WEDNESDAY and \
                     day.month == month][2]
    return imm_date 


@dataclass
class IRFuture():
    effective_date: Optional[pd.Timestamp]=None
    maturity_date: Optional[pd.Timestamp]=None
    price: Optional[float] = None
    futures_interest_rate: Optional[float] = None
    price_date: Optional[pd.Timestamp] = None
    convexity: float = 0.0

    day_count_basis: InitVar[Optional[VALID_DAY_COUNT_BASIS]] = None
    imm_delivery_year: InitVar[Optional[int]] = None
    imm_delivery_month: InitVar[Optional[int]] = None
    
    daycounter: DayCounter = field(init=False)
    forward_interest_rate: float = field(init=False)
    

    def __post_init__(self, day_count_basis, imm_delivery_year, imm_delivery_month):
        
        self.daycounter = DayCounter(day_count_basis)
        
        if self.effective_date is None and self.maturity_date is None \
            and self.imm_delivery_year is not None and self.imm_delivery_month is not None:
            maturity_year = imm_delivery_year
            maturity_month = imm_delivery_month
            effective_month = 12 if maturity_month == 3 else maturity_month - 3
            effective_year = maturity_year - 1 if maturity_month == 3 else maturity_year 
            
            self.effective_date = getIMMDate(effective_year,effective_month)
            self.maturity_date = getIMMDate(maturity_year,maturity_month)
        
        #if not isIMMDate(self.effective_date): warnings.warn("UserWarning: " + "effective_date " + self.effective_date.strftime('%Y-%m-%d-%a') + " is not an IMM date",UserWarning)
        #if not isIMMDate(self.maturity_date): warnings.warn("UserWarning: " + "maturity_date " + self.maturity_date.strftime('%Y-%m-%d-%a') + " is not an IMM date",UserWarning)
        if self.price is None and self.futures_interest_rate is None: raise ValueError("'price' or 'futures_interest_rate' must be specified")
        if self.effective_date >= self.maturity_date: raise ValueError("'effective_date' is equal to or after the 'maturity_date'")
        if self.price_date is not None and self.price_date > self.maturity_date: raise ValueError("'price_date' is equal to or after the 'maturity_date'")
        if self.price is None: self.price = 100 - self.futures_interest_rate * 100
        if self.futures_interest_rate is None: self.futures_interest_rate = (100 - self.price) / 100
        
        self.forward_interest_rate = self.futures_interest_rate + self.convexity    
        

@dataclass
class FRA():
    contract_rate: float
    transaction_date: pd.Timestamp  
    forward_starting: int = 2
    spot_date: Optional[pd.Timestamp]=None  
    wait_period: Optional[int]=None 
    contract_period: Optional[int]=None 
    effective_date: Optional[pd.Timestamp]=None
    maturity_date: Optional[pd.Timestamp]=None
    currency_holidays: Optional[VALID_CCYS]=None
    city_holidays: Optional[str]=None
    
    day_count_basis: InitVar[Optional[VALID_DAY_COUNT_BASIS]] = None
    fra_str: InitVar[Optional[str]] = None
    holiday_calendar: InitVar[Optional[np.busdaycalendar]]=None
    
    def __post_init__(self, day_count_basis, fra_str, holiday_calendar):

        if holiday_calendar is None: 
            holiday_calendar = get_calendar(self.currency_holidays, self.city_holidays)         

        if self.spot_date is None: 
            date = np.datetime64(self.transaction_date + pd.DateOffset(days=self.forward_starting),'D')
            np.busday_offset(dates=date, offsets=0, roll='following', busdaycal=holiday_calendar)
            
        if fra_str is not None:
            contract_period, wait_period = fra_str.upper().split('X')
            contract_period = int(contract_period)
            wait_period = int(wait_period)
        
        if self.effective_date is None and self.maturity_date is None:
            self.effective_date = self.spot_date + pd.DateOffset(months=self.wait_period) 
            self.maturity_date = self.spot_date + pd.DateOffset(months=self.contract_period)
                    
        self.daycounter = DayCounter(day_count_basis)





# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())


import calendar
import datetime as dt
import numpy as np
import pandas as pd
from dataclasses import dataclass
from typing import Optional, Literal, Union

# VALID_DAY_COUNT_BASIS_TYPES = Union[Literal['30/360'],
#                               Literal['30e/360'], 
#                               Literal['30e/360 isda'], 
#                               Literal['act/360'], 
#                               Literal['act/365'], 
#                               Literal['act/act']]

VALID_DAY_COUNT_BASIS = ['30/360','30e/360','30e/360_isda','act/360','act/365','act/act']
VALID_DAY_COUNT_BASIS_TYPES = Union[tuple(Literal[i] for i in VALID_DAY_COUNT_BASIS)]

@dataclass
class DayCounter:
    day_count_basis: Optional[VALID_DAY_COUNT_BASIS_TYPES] = 'act/act'
    VALID_DAY_COUNT_BASIS = VALID_DAY_COUNT_BASIS 
    
    def __post_init__(self):
        if self.day_count_basis in [None, np.nan, '']:
            self.day_count_basis = 'act/act' 
        elif self.day_count_basis not in self.VALID_DAY_COUNT_BASIS:
            raise ValueError(f"Invalid day_count_basis. Must be one of {VALID_DAY_COUNT_BASIS}")
           
    def to_datetime_or_datetimeindex(self, date_object) -> 'dt.datetime, pd.DatetimeIndex':
        if type(date_object) == pd.Timestamp:
            return date_object.to_pydatetime()
        elif type(date_object) == np.datetime64:
            return pd.to_datetime(date_object)
        elif type(date_object) == dt.date:
            return dt.datetime(date_object.year,date_object.month,date_object.day)
        elif type(date_object) == pd.Series or type(date_object) == list:
            return pd.DatetimeIndex(date_object)
        else:
            return date_object


    def clean_up(self, start_date, end_date):
        start_date = self.to_datetime_or_datetimeindex(start_date)
        end_date = self.to_datetime_or_datetimeindex(end_date)
        
        if type(start_date) == dt.datetime and type(end_date) == pd.DatetimeIndex:
            start_date = pd.DatetimeIndex([start_date for _ in range(len(end_date))])
        elif type(start_date) == pd.DatetimeIndex and type(end_date) == dt.datetime:
            end_date = pd.DatetimeIndex([end_date for _ in range(len(start_date))])
        
        return start_date, end_date

    def day_count(self, 
                  start_date,
                  end_date, 
                  is_end_date_on_termination: bool=False):
        
        start_date, end_date = self.clean_up(start_date, end_date)
        
        if self.day_count_basis in {'act/360','act/365','act/act'}:
            return (end_date - start_date).days
        
        elif self.day_count_basis == '30/360':
            d1 = np.minimum(30, start_date.day)
            
            # these lines need to work for dates and datetimeindex
            #d2 = np.minimum(d1, end_date.day) if d1 == 30 else end_date.day
            d2 = end_date.day.values
            idx = d1 == 30
            d2[idx] = np.minimum(d1, end_date.day).values[idx]
                 
            return 360*(end_date.year - start_date.year)\
                   + 30*(end_date.month - start_date.month)\
                   + d2 - d1
        
        elif self.day_count_basis == '30e/360':
            d1 = np.minimum(30, pd.Series(start_date.day))
            d2 = np.minimum(30, pd.Series(end_date.day))
            return 360 * (pd.Series(end_date.year) - pd.Series(start_date.year)) \
                   + 30 * (pd.Series(end_date.month) - pd.Series(start_date.month)) \
                   + d2 - d1
                   

        elif self.day_count_basis == '30e/360 isda':
    
            def _is_last_day_of_feb(date):
                if date.month == 2:
                    _, last_of_month = calendar.monthrange(date.year, date.month)
                    if date.day == last_of_month:
                        return True
                return False

            if start_date.day == 31 or _is_last_day_of_feb(start_date):
                d1 = 30
            else:
                d1 = start_date.day

            if end_date.day == 31 or (_is_last_day_of_feb(end_date)
                                     and not is_end_date_on_termination):
                d2 = 30
            else:
                d2 = end_date.day

            return 360 * (end_date.year - start_date.year) \
                   + 30 * (end_date.month - start_date.month) \
                   + d2 - d1




    def year_fraction(self, 
                      start_date,
                      end_date,
                      is_end_date_on_termination: bool=False):
        
       

        if self.day_count_basis in {'30/360','30e/360','30e/360 isda', 'act/360'}:
            return self.day_count(start_date, end_date, is_end_date_on_termination) / 360.0
        
        elif self.day_count_basis == 'act/act':
            start_date, end_date = self.clean_up(start_date, end_date)
            
            start_year = pd.Series(start_date.year)
            end_year = pd.Series(end_date.year)
            year_1_diff = 365 + start_year.map(calendar.isleap)
            year_2_diff = 365 + end_year.map(calendar.isleap) 
            
            total_sum = end_year - start_year - 1
            diff_first = pd.Series([dt.datetime(v + 1, 1, 1) for v in start_year]) - start_date
            total_sum += diff_first.dt.days / year_1_diff
            diff_second = end_date - pd.Series([dt.datetime(v, 1, 1) for v in end_year]) 
            total_sum += diff_second.dt.days / year_2_diff

            if type(start_date) is pd.DatetimeIndex:
                return total_sum
            elif type(start_date) is dt.datetime:
                return total_sum[0]

        elif self.day_count_basis == 'act/365':
            return self.day_count(start_date, end_date, is_end_date_on_termination) / 365.0
        


import xlsxwriter

def write_trade(path,cashflow_schedules,DV01):

    def write_leg(df, row, dynamic=True):
        df = df.dropna()
        
        # Write the header of the cashflow dataframe
        col_names = df.columns.values
        for c,v in enumerate(col_names):
            worksheet.write(row,c,v,header_format)
        row +=  1
        
        # Write the cashflow dataframe values
        arr = df.values
        nb_rows = len(arr)
        nb_cols = len(arr[0])
        map_colnb_to_colname = {i : col_name for i,col_name in enumerate(col_names)}
        worksheet.set_column(0, 12, 14) # Set column width of colums 0-12 to 14    
        
        # Use formula's for the net floating rate, cash flow & cash flow PV columns
        if dynamic: 
            hardcoded_cols = {'Start Date', 'End Date', 'Payment Date', 'Days', 'Year Fraction', 'Notional', 'Fixed Rate', 'Fixing Date', 'Index Rate','Spread','Discount Factor'}
            for r in range(nb_rows):
                for c,col_name in enumerate(col_names):
                    if col_name in hardcoded_cols:
                        worksheet.write(row,c,arr[r,c],format_dict[map_colnb_to_colname[c]])
                    else:
                        if nb_cols == 10:
                            if col_name == 'Cash Flow':
                                worksheet.write(row,c,'=E' + str(row+1)+'*F'+str(row+1)+'*G'+str(row+1),format_dict[map_colnb_to_colname[c]])
                            elif col_name == 'Cash Flow PV':
                                worksheet.write(row,c,'=H' + str(row+1)+'*I'+str(row+1),format_dict[map_colnb_to_colname[c]])
                            else:
                                raise ValueError('Invalid value: ', col_name, nb_cols)
                        elif nb_cols == 13:
                            if col_name == 'Floating Rate':
                                worksheet.write(row,c,'=H' + str(row+1)+'+I'+str(row+1),format_dict[map_colnb_to_colname[c]])
                            elif col_name == 'Cash Flow':
                                worksheet.write(row,c,'=E' + str(row+1)+'*F'+str(row+1)+'*J'+str(row+1),format_dict[map_colnb_to_colname[c]])
                            elif col_name == 'Cash Flow PV':
                                worksheet.write(row,c,'=K' + str(row+1)+'*L'+str(row+1),format_dict[map_colnb_to_colname[c]])
                            else:
                                raise ValueError('Invalid value: ', col_name, nb_cols)
                row +=  1
                                
        # Hardcode data  
        else:
            for r in range(nb_rows):
                for c in range(nb_cols):
                    worksheet.write(row,c,arr[r,c],format_dict[map_colnb_to_colname[c]])
                row +=  1
        
        # Write formula to sum the discounted cashflows
        col_letter = str(xlsxwriter.utility.xl_col_to_name(nb_cols-1))
        worksheet.write(row,nb_cols-1,'=SUM(' + col_letter + str(row-nb_rows+1) + ':' + col_letter + str(row) + ')',sum_format)
        cell = str(xlsxwriter.utility.xl_col_to_name(nb_cols-1)) + str(row+1)
        row += 2
        
        return row, cell
    
    
    def write_cashflows(worksheet, dynamic_flag):
        row = 0
        cells = []
        for i,leg in enumerate(cashflow_schedules):
            row,cell = write_leg(cashflow_schedules[i], row, dynamic_flag)
            cells.append(cell)
            
        worksheet.write(row,1,"MV",header_format)
        worksheet.write(row,2,"DV01",header_format)
        row += 1
        for i,cell in enumerate(cells):
            worksheet.write(row,0,"LEG"+str(i+1),number_format)
            worksheet.write(row,1,"="+cell,number_format)
            worksheet.write(row,2,DV01[i],number_format)
            row += 1
            
        worksheet.write(row,0,"Net")
        worksheet.write(row,1,"=SUM(" + "B" + str(row - len(cells) + 1) + ":B" + str(row) + ')',sum_format)
        worksheet.write(row,2,"=SUM(" + "C" + str(row - len(cells) + 1) + ":C" + str(row) + ')',sum_format)
    
        
    workbook = xlsxwriter.Workbook(path)
    
    date_format = workbook.add_format({'num_format': 'd-mmm-yyyy;;'})
    number_format = workbook.add_format({'num_format': '#,##0;(#,##0)'})
    fraction_format = workbook.add_format({'num_format': '0.0000_-'})
    percentage_format = workbook.add_format({'num_format': '0.0000%'})
    general_format = workbook.add_format({'num_format': 'General'})
    header_format = workbook.add_format()
    header_format.set_bottom()
    sum_format = workbook.add_format({'num_format': '#,##0;(#,##0)'})
    sum_format.set_top()
    sum_format.set_bottom()
    
    format_dict = {'Fixing Date' : date_format,
                   'Start Date' : date_format,
                   'End Date' : date_format,
                   'Payment Date' : date_format,
                   'Days' : general_format,
                   'Year Fraction' : fraction_format,
                   'Notional' : number_format,
                   'Fixed Rate' : percentage_format,
                   'Index Rate' : percentage_format,
                   'Spread' : percentage_format,
                   'Floating Rate' : percentage_format,
                   'Cash Flow' : number_format,
                   'Discount Factor' : fraction_format,
                   'Cash Flow PV' : number_format}
    
    worksheet = workbook.add_worksheet('Hardcoded')
    write_cashflows(worksheet, dynamic_flag=False)
    worksheet = workbook.add_worksheet('Dynamic')
    write_cashflows(worksheet, dynamic_flag=True)
    
    workbook.close()
    




# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    
from frm.frm.schedule.business_day_calendar import get_calendar #, VALID_CITY_HOLIDAYS, VALID_CURRENCY_HOLIDAYS

import numpy as np
import pandas as pd
import datetime as dt
from typing import Literal, Optional


VALID_DAY_ROLL = Literal[tuple(range(1,32)),'EOM']
VALID_LAST_STUB = Literal['short','long']
VALID_PAYMENT_TYPE = Literal['in_arrears','in_advance']
VALID_PAYMENT_FREQUENCY = Literal['w','28d','m','q','s','a','z', \
                                  'weekly','28days','monthly','quarterly','semiannually','annually','zerocoupon']
VALID_PERIOD_TYPES = Literal['days','weeks','months','years']
VALID_ROLL_CONVENTION = Literal['actual','following','preceding','modifiedfollowing','modifiedpreceding']
VALID_STUB = Literal['short','long']
VALID_STUB_GENERAL = Literal['first_short','first_long','last_short','last_long']

nb_periods_period_type = {'w':   (1,'weeks'), 'weekly':   (1,'weeks'),
                          '28d': (4,'weeks'), '28days':   (4,'weeks'),
                          'm':   (1,'months'), 'monthly': (1,'months'),
                          'q':   (3,'months'), 'quarterly': (3,'months'),
                          's':   (6,'months'), 'semiannually': (6,'months'),
                          'a':   (1,'years'), 'annually': (1,'years')}
                          

def roll_dates(dates: pd.DatetimeIndex, 
               day_roll: VALID_DAY_ROLL) -> pd.DatetimeIndex:
    """Roll the days of the dates to the provided day_roll."""
    if str(day_roll).upper() == 'EOM' or day_roll == 31:
        return dates + pd.offsets.MonthEnd(0)
    elif day_roll < 29:
        return pd.DatetimeIndex([dt.datetime(d.year,d.month,int(day_roll)) for d in dates])
    else:
        rolled = []
        for d in dates:
            try:
                rolled.append(dt.datetime(d.year,d.month,int(day_roll)))
            except ValueError:
                rolled.append(d + pd.offsets.MonthEnd(0))
        return pd.DatetimeIndex(rolled)     


def date_offset(nb_periods: int,
                period_type: VALID_PERIOD_TYPES) -> pd.DateOffset:
    if period_type == 'months':
        return pd.DateOffset(months=nb_periods)
    elif period_type == 'years':
        return pd.DateOffset(years=nb_periods)
    elif period_type == 'weeks': 
        return pd.DateOffset(weeks=nb_periods)
    elif period_type == 'days':
        return pd.DateOffset(days=nb_periods)
    
    
def payment_schedule(start_date: pd.Timestamp,
                     end_date: pd.Timestamp,
                     payment_freq: VALID_PAYMENT_FREQUENCY,
                     roll_convention: Optional[VALID_ROLL_CONVENTION]='modifiedfollowing',
                     day_roll: Optional[VALID_DAY_ROLL]=None,
                     stub: Optional[VALID_STUB_GENERAL]='first_short',
                     first_stub: Optional[VALID_STUB]=None,
                     last_stub: Optional[VALID_STUB]=None,
                     first_cpn_end_date: Optional[pd.Timestamp]=None,
                     last_cpn_start_date: Optional[pd.Timestamp]=None,
                     payment_type: Optional[VALID_PAYMENT_TYPE]='in_arrears', 
                     payment_delay: int=0,
                     add_fixing_dates: bool=False,
                     add_initial_exchange_period: bool=False,
                     fixing_days_ahead: int=2,
                     currency_holidays=None,
                     city_holidays=None,
                     holiday_cal: np.busdaycalendar=None) -> pd.DataFrame:
    """
    Create a payment schedule.

    Parameters
    ----------
    start_date : pandas.Timestamp
        Specifies the effective date of the schedule
    end_date : pandas.Timestamp
        Specifies the expiration date of the schedule
    payment_freq : {'W','28D','M', 'Q', 'S', 'A'}
        Specify the payment frequency
    day_count_basis : {'ACT/ACT','ACT/360','ACT/365', '30/360', '30E/360', '30E/350 ISDA'}
    roll_convention : {'actual','following','preceding','modifiedfollowing','modifiedpreceding'},
        How to treat dates that do not fall on a valid day. The default is ‘raise’.
            'following' means to take the first valid day later in time.
            'preceding' means to take the first valid day earlier in time.
            'modifiedfollowing' means to take the first valid day later in time unless it is across a Month boundary, in which case to take the first valid day earlier in time.
            'modifiedpreceding' means to take the first valid day earlier in time unless it is across a Month boundary, in which case to take the first valid day later in time.
    stub : {'first short','first long','last short','last long'}
        Specify the type and location of stub period
    first_stub : {short, long}
        Specifies the type of the first stub
    last_stub : {short, long}
        Specifies the type of the last stub
    payment_type : {'in arrears','in advance'}
        Specifies when payments are made
    payment_delay : int
        Specifies how many days after period end the payment is made
    add_fixing_dates : bool
    fixing_days_ahead : int
        Specifies the number of days before the peiord start date the fixing date occur
    day_roll : {1,2,3,...,30,31,'EOM'}
        Specifies the day periods should start/end on, EOM = roll to the end of month
    currency_holidays : array of strings of currency code of the locale to observe
    city_holidays : array of strings of the city holidays to observe {'Amsterdam','AMS',
        Specifies the holidays of the countrys and finp.nancial centres to be observed
    holiday_cal : np.busdaycalendar 
        Specifies the holiday calendar to observe. If this parameter is provided, neither ccy or holidays may be provided
        
    Returns
    -------
    schedule : pandas.DataFrame
        Columns
            start_date 
            end_date 
            payment_date 
    """
        
    if roll_convention is None: roll_convention = 'modifiedfollowing'
    if payment_type is None: payment_type = 'in_arrears'
    if payment_delay is None: payment_delay = 0
    
    payment_freq = payment_freq.lower()
    roll_convention = roll_convention.lower()
    payment_type = payment_type.lower().replace(' ','')
    
    nb_periods, period_type = nb_periods_period_type[payment_freq.lower()]
    
    if holiday_cal is None: holiday_cal = get_calendar(currency_holidays, city_holidays) 
        
    if first_cpn_end_date is not None:
        # The expected first coupon end date if no stub
        expected_first_cpn_end_date = pd.Timestamp(np.busday_offset(dates=pd.DatetimeIndex([start_date - date_offset(nb_periods, period_type)]).astype(str), offsets=0, roll=roll_convention, busdaycal=holiday_cal)[0])
    if last_cpn_start_date is not None:
        # The expected last coupon start date if no stub
        expected_last_cpn_start_date = pd.Timestamp(np.busday_offset(dates=pd.DatetimeIndex([end_date - date_offset(nb_periods, period_type)]).astype(str), offsets=0, roll=roll_convention, busdaycal=holiday_cal)[0])

    # If the schedule terms specifiy 1 period
    if (first_cpn_end_date is not None and first_cpn_end_date == end_date) or \
        (last_cpn_start_date is not None and last_cpn_start_date == start_date) or \
        payment_freq in {'z','zc','zerocoupon'}:
        d1 = [start_date]
        d2 = [end_date]
        
    # If the schedule terms specifiy 2 periods
    elif first_cpn_end_date is not None and last_cpn_start_date is not None and (first_cpn_end_date == last_cpn_start_date):
        d1 = [start_date, first_cpn_end_date]
        d2 = [last_cpn_start_date, end_date]
        
    # Otherwise, calculate the schedule dates
    else:
        # If the trade has a maximum of one stub or if the first_cpn_end_date or first_cpn_end_date is specified, 
        # the date is where we expect it to be without a stub
        if (first_cpn_end_date is None and last_cpn_start_date is None) \
            or (first_cpn_end_date is not None and first_cpn_end_date == expected_first_cpn_end_date) \
            or (last_cpn_start_date is not None and last_cpn_start_date == expected_last_cpn_start_date):
                
            # Use a first short stub if:
            # 1) specified
            # 2) no stub parameters (stub, last_stub, first_cpn_end_date, last_cpn_start_date) where provided
            # 3) the last_cpn_start_date is in the expected position if there is no last stub
            if stub == 'first_short' or first_stub == 'short' \
               or stub is None and first_stub is None and last_stub is None and first_cpn_end_date is None and last_cpn_start_date is None \
               or (last_cpn_start_date is not None and last_cpn_start_date == expected_last_cpn_start_date):
                d1,d2 = backward_date_generation(start_date, end_date, nb_periods, period_type)       
                                                         
        
            # Use a last short stub if:
            # 1) specified
            # 2) the first_cpn_end_date is in the expected position if there is no first stub
            elif stub =='last_short' or last_stub == 'short' \
                or (first_cpn_end_date is not None and first_cpn_end_date == expected_first_cpn_end_date):
                d1,d2 = forward_date_generation(start_date, end_date, nb_periods, period_type) 
               
            
            # Use a first long stub only when specified
            elif stub == 'first_long' or first_stub == 'long':
                d1,d2 = backward_date_generation(start_date, end_date, nb_periods, period_type) 
                d1 = [d1[0]] + d1[2:]
                d2 = d2[1:]
                
            
            # Use a last long stub only when specified
            elif stub == 'last_long' or last_stub == 'long':
                d1,d2 = forward_date_generation(start_date, end_date, nb_periods, period_type) 
                d1 = d1[:-1]  
                d2 = d2[:-2] + [d2[-1]]
               
        # If there is a stub at both the start and end
        elif first_cpn_end_date is not None and last_cpn_start_date is not None:
            d1,d2 = forward_date_generation(first_cpn_end_date, last_cpn_start_date, nb_periods, period_type) 
            d1 = [start_date] + d1 + [last_cpn_start_date]
            d2 = [first_cpn_end_date] + d2 + [end_date]
   
            
        # dates will need to be rolled to the start date day
        elif first_cpn_end_date is None and last_cpn_start_date is not None:
            d1,d2 = backward_date_generation(start_date, last_cpn_start_date, nb_periods, period_type) 
            d1 = d1 + [last_cpn_start_date]
            d2 = d2 + [end_date]
          
            
        # dates will need to be rolled to the end date day
        elif first_cpn_end_date is not None and last_cpn_start_date is None:
            d1,d2 = forward_date_generation(first_cpn_end_date, end_date, nb_periods, period_type) 
            d1 = [start_date] + d1
            d2 = [first_cpn_end_date] + d2
            
              
        d1 = pd.DatetimeIndex(d1)
        d2 = pd.DatetimeIndex(d2)
              
        if day_roll is not None:
            d1 = roll_dates(d1, day_roll)
            d2 = roll_dates(d2, day_roll)
        
        if roll_convention != 'actual':
            # Roll the days of the schedule per the roll convention and business day holiday calendar
            d1 = pd.DatetimeIndex(np.busday_offset(dates=d1.values.astype('datetime64[D]'), offsets=0, roll=roll_convention, busdaycal=holiday_cal))
            d2 = pd.DatetimeIndex(np.busday_offset(dates=d2.values.astype('datetime64[D]'), offsets=0, roll=roll_convention, busdaycal=holiday_cal))

        # Use the dates specified in the function call (start_date, end_date, first_cpn_end_date, last_cpn_end_date) 
        # in the schedule, regardless if they fall on a non-business day. 
        d1 = d1.to_list()
        d2 = d2.to_list()
        if first_cpn_end_date is not None and last_cpn_start_date is not None:
            d1 = [start_date] + [first_cpn_end_date] + d1[2:-1] + [last_cpn_start_date]
            d2 = [first_cpn_end_date] + d2[1:-2] + [last_cpn_start_date] + [end_date]
        elif first_cpn_end_date is None and last_cpn_start_date is not None:
            d1 = [start_date] + d1[1:-1] + [last_cpn_start_date]
            d2 = d2[:-2] + [last_cpn_start_date] + [end_date]        
        elif not first_cpn_end_date is None and last_cpn_start_date is None:  
            d1 = [start_date] + [first_cpn_end_date] + d1[2:]
            d2 = [first_cpn_end_date] + d2[1:-1] + [end_date]
        else: 
            d1 = [start_date] + d1[1:]
            d2 = d2[:-1] + [end_date]
                
    df = pd.DataFrame({'period_start': d1, 'period_end': d2})
    df = df[df['period_start'] != df['period_end']]
    df.reset_index(drop=True,inplace=True)

    # Add the payment dates
    if payment_type == 'in_arrears':
        if payment_delay == 0:
            df['payment_date'] = df['period_end']
        else:
            df['payment_date'] = np.busday_offset(pd.DatetimeIndex(df['period_end']+pd.DateOffset(days=payment_delay)).values.astype('datetime64[D]'), offsets=0, roll='following', busdaycal=holiday_cal)
    elif payment_type == 'in_advance':
        if payment_delay == 0:
            df['payment_date'] = df['period_start'] 
        else:
            df['payment_date'] = np.busday_offset(pd.DatetimeIndex(df['period_start']+pd.DateOffset(days=payment_delay)).values.astype('datetime64[D]'), offsets=0, roll='following', busdaycal=holiday_cal)  

    # Add the fixing dates
    if add_fixing_dates:
        if fixing_days_ahead == 0:
            df['fixing_period_start'] = df['period_start'] 
            df['fixing_period_end'] = df['period_end'] 
        else:
            df['fixing_period_start'] = np.busday_offset(pd.DatetimeIndex(df['period_start']-pd.DateOffset(days=fixing_days_ahead)).values.astype('datetime64[D]'), offsets=0, roll='preceding', busdaycal=holiday_cal)
            df['fixing_period_end'] = np.busday_offset(pd.DatetimeIndex(df['period_end']-pd.DateOffset(days=fixing_days_ahead)).values.astype('datetime64[D]'), offsets=0, roll='preceding', busdaycal=holiday_cal)
        df = df[['fixing_period_start','fixing_period_end','period_start','period_end','payment_date']]

    # Add initial notional exchange date
    if add_initial_exchange_period:
        payment_date = np.busday_offset(pd.DatetimeIndex([start_date+pd.DateOffset(days=payment_delay)]).values.astype('datetime64[D]'), offsets=0)
        row = {'period_start': start_date,'period_end': start_date,'payment_date':payment_date}
        df = pd.concat([pd.DataFrame(row), df], ignore_index=True)

    return df
    

def forward_date_generation(start_date: pd.Timestamp, 
                            end_date: pd.Timestamp, 
                            nb_periods: int,
                            period_type: VALID_PERIOD_TYPES) -> (list, list):
    """
    Generates a schedule working forwards from start date to the end date

    Parameters
    ----------
    start_date : pd.Timestamp
        start date of the schedule
    end_date : pd.Timestamp
        end date of the schedule
    nb_periods : Integer
        Specifies the number of periods of length 'period_type' per time increment
    period_type : {'years','months','weeks','days'}
        Specifies the length of one period

    Returns
    -------
    tuble of lists
        start dates, end dates
    """

    d1_arr = []
    d2_arr = []
    d1 = start_date
    d2 = start_date + date_offset(nb_periods,period_type)
    i = 1

    while end_date > d2 :   
        d1_arr.append(d1)
        d2_arr.append(d2)
        d1 = start_date + date_offset(nb_periods*i,period_type) 
        d2 = start_date + date_offset(nb_periods*(i+1),period_type) 
        i = i + 1
    d1_arr.append(d1)
    d2_arr.append(end_date)

    return d1_arr,d2_arr


def backward_date_generation(start_date: pd.Timestamp, 
                             end_date: pd.Timestamp, 
                             nb_periods: int, 
                             period_type: VALID_PERIOD_TYPES) -> (list, list):
    """
    Generates a schedule working backwards from the end date to the start date

    Parameters
    ----------
    start_date : pd.Timestamp
        start date of the schedule
    end_date : pd.Timestamp
        end date of the schedule
    nb_periods : Integer
        Specifies the number of periods of length 'period_type' per time decrement
    period_type : {'years','months','weeks','days'}
        Specifies the length of one period

    Returns
    -------
    tuple of lists
        start dates, end dates
    """
    
    d1_arr = []
    d2_arr = []
    d1 = end_date - date_offset(nb_periods,period_type)
    d2 = end_date
    i = 1
    while start_date < d1:   
        d1_arr.append(d1)
        d2_arr.append(d2)
        d1 = end_date - date_offset(nb_periods*(i+1),period_type) 
        d2 = end_date - date_offset(nb_periods*i,period_type) 
        i = i + 1
    d1_arr.append(start_date)
    d2_arr.append(d2)

    return list(reversed(d1_arr)),list(reversed(d2_arr))


def create_date_grid_for_fx_exposures(curve_date: pd.Timestamp, 
                                      delivery_dates: np.array, 
                                      sampling_freq: str=None,
                                      date_grid_pillar: pd.DatetimeIndex=None,
                                      payments_on_value_date_have_value: bool=False,
                                      include_last_day_of_value: bool=False,
                                      include_day_after_last_day_of_value: bool=False) -> pd.DatetimeIndex:
    """
    Defines a date grid for calculating exposures
    Please note, for each trade, this defines the settlement date grid, hence the expiry
    date grid needs to be solved based on the delay of the market.
    For FX derivatives this is relively simple as market data inputs can be interploted from
    (i) expiry or (ii) settle dates.

    Parameters
    ----------
    curve_date : pd.Timestamp
        DESCRIPTION.
    delivery_dates : np.array
        DESCRIPTION.
    sampling_freq : str, optional
        DESCRIPTION. The default is None.
    date_grid_pillar : pd.DatetimeIndex, optional
        DESCRIPTION. The default is None.
    include_payments_on_value_date : bool, optional
        DESCRIPTION. The default is False.
    include_last_day_of_value : bool, optional
        DESCRIPTION. The default is False.
    include_day_after_last_day_of_value : bool, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    date_grid : TYPE
        DESCRIPTION.

    """
        
    # Challenge is that exposures are defined using the rate-fixing - i.e the expiry date
    # The expiry dates are what we need to specify for our FX option valuation or FX rate simulation
    # The trade technically still has value until some time on day of delivery/settlement.
    # Hence the date grid should be constructed on a delivery/settlement basis. 
    # There is a pototo/potata view on if a cashflow has value on the settlement date. 
    # Our default view is no, it should be in accounts receivable, but this can be toggled. 
    
    delivery_dates = np.unique(delivery_dates) 
    max_settlement_date = delivery_dates.max()
    
    if sampling_freq is not None:    
        if sampling_freq == '1d':
            date_grid = forward_date_generation(curve_date, max_settlement_date, 1, 'days')
        elif sampling_freq == '1w':
            date_grid = forward_date_generation(curve_date, max_settlement_date, 1, 'weeks')
        elif sampling_freq == '1m':
            date_grid = forward_date_generation(curve_date, max_settlement_date, 1, 'months')
        elif sampling_freq == '3m':
            date_grid = forward_date_generation(curve_date, max_settlement_date, 3, 'months')    
        elif sampling_freq == '6m':
             date_grid = forward_date_generation(curve_date, max_settlement_date, 6, 'months')   
        elif sampling_freq == '12m':
             date_grid = forward_date_generation(curve_date, max_settlement_date, 12, 'months') 
             
        date_grid = pd.DatetimeIndex([curve_date]).append(pd.DatetimeIndex(date_grid[1]))
        
    elif date_grid_pillar is not None:
        assert isinstance(date_grid_pillar, pd.DatetimeIndex) 
        date_grid = date_grid_pillar
    else:
        raise ValueError("Either 'sampling_freq' or date_grid_pillar' must be specified")

    if include_last_day_of_value:
        if payments_on_value_date_have_value:
            last_day_of_value = delivery_dates
        else: 
            last_day_of_value = delivery_dates - pd.DateOffset(days=1)
        date_grid = pd.DatetimeIndex(np.concatenate(date_grid.values, last_day_of_value))

    elif include_day_after_last_day_of_value:
        if payments_on_value_date_have_value:
            last_day_of_value = delivery_dates + pd.DateOffset(days=1)
        else: 
            last_day_of_value = delivery_dates 
        date_grid = pd.DatetimeIndex(np.concatenate(date_grid.values, last_day_of_value))

    date_grid = date_grid.drop_duplicates().sort_values()
    return date_grid




# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    
from frm.frm.schedule.daycounter import DayCounter, VALID_DAY_COUNT_BASIS
from frm.frm.schedule.schedule import payment_schedule, VALID_DAY_ROLL, VALID_LAST_STUB, VALID_PAYMENT_TYPE, VALID_PAYMENT_FREQUENCY, VALID_STUB, VALID_ROLL_CONVENTION
from frm.frm.schedule.business_day_calendar import get_calendar
from frm.frm.instruments.ir.leg import Leg

import numpy as np
import pandas as pd
from scipy.optimize import fsolve
from typing import Literal, Optional
        
                                                       
class Swap():    
    def __init__(self,
                 **kwargs # dictionary of swap term defaults 
                 ):
        
        if kwargs != {}:
            dict_default_terms = kwargs['default_terms']
            
            tmp = {}
            for k,v in dict_default_terms.items():
                if k[:4] != 'pay_' and k[:4] != 'rec_':
                    tmp['pay_'+k] = v
                    tmp['rec_'+k] = v

        dict_init = {**dict_default_terms, **tmp, **kwargs['specified_terms']}
        
        list_all_inputs = ['effective_date',
                        'maturity_date',
                        'tenor',
                        'transaction_date',
                        'forward_starting',
                        'tenor_calendar',
                        'tenor_city_holidays',
                        'tenor_currency_holidays',
                         # Pay Schedule
                        'pay_payment_frequency',
                        'pay_roll_convention',
                        'pay_day_roll',
                        'pay_stub',
                        'pay_first_stub',
                        'pay_last_stub',
                        'pay_first_cpn_end_date',
                        'pay_last_cpn_start_date',
                        'pay_payment_type',
                        'pay_payment_delay',
                        'pay_fixing_days_ahead',
                        'pay_currency_holidays',
                        'pay_city_holidays',
                        'pay_holiday_calendar',
                         # Pay Notional & Currencies
                        'pay_notional',
                        'pay_notional_currency',
                        'pay_settlement_currency',
                        'pay_exchange_notionals',
                        'pay_MTM_notional_currency',
                        'pay_MTM_notional',
                         # Pay cpn Details
                        'pay_leg_type',
                        'pay_fixed_rate',
                        'pay_float_spread',
                        'pay_cpn_cap',
                        'pay_cpn_floor',
                        'pay_day_count_basis',
                         # Pay pricing yield curves
                        'pay_discount_curve',
                        'pay_forward_curve',
                        '',
                         # Rec Schedule
                        'rec_payment_frequency',
                        'rec_roll_convention',
                        'rec_day_roll',
                        'rec_stub',
                        'rec_first_stub',
                        'rec_last_stub',
                        'rec_first_cpn_end_date',
                        'rec_last_cpn_start_date',
                        'rec_payment_type',
                        'rec_payment_delay',
                        'rec_fixing_days_ahead',
                        'rec_currency_holidays',
                        'rec_city_holidays',
                        'rec_holiday_calendar',
                         # Rec Notional & Currencies
                        'rec_notional',
                        'rec_notional_currency',
                        'rec_settlement_currency',
                        'rec_exchange_notionals',
                        'rec_MTM_notional_currency',
                        'rec_MTM_notional',
                         # Rec cpn Details
                        'rec_leg_type',
                        'rec_fixed_rate',
                        'rec_float_spread',
                        'rec_cpn_cap',
                        'rec_cpn_floor',
                        'rec_day_count_basis',
                         # Rec pricing yield curves
                        'rec_discount_curve',
                        'rec_forward_curve',
                        ]      
        
        for k in list_all_inputs:
            if k not in dict_init.keys():
                dict_init[k] = None
   
        # Swap tenor 
        effective_date = dict_init['effective_date']
        maturity_date = dict_init['maturity_date']
        tenor = dict_init['tenor']
        transaction_date = dict_init['transaction_date']
        forward_starting = dict_init['forward_starting']
        # Pay Schedule
        pay_payment_frequency = dict_init['pay_payment_frequency']
        pay_roll_convention = dict_init['pay_roll_convention']
        pay_day_roll = dict_init['pay_day_roll']
        pay_stub = dict_init['pay_stub']
        pay_first_stub = dict_init['pay_first_stub']
        pay_last_stub = dict_init['pay_last_stub']
        pay_first_cpn_end_date = dict_init['pay_first_cpn_end_date']
        pay_last_cpn_start_date = dict_init['pay_last_cpn_start_date']
        pay_payment_type = dict_init['pay_payment_type']
        pay_payment_delay = dict_init['pay_payment_delay']
        pay_fixing_days_ahead = dict_init['pay_fixing_days_ahead']
        pay_currency_holidays = dict_init['pay_currency_holidays']
        pay_city_holidays = dict_init['pay_city_holidays']
        pay_holiday_calendar = dict_init['pay_holiday_calendar']
        # Pay Notional & Currencies
        pay_notional = dict_init['pay_notional']
        pay_transaction_price = dict_init['pay_notional']
        pay_notional_currency = dict_init['pay_notional_currency']
        pay_settlement_currency = dict_init['pay_settlement_currency']
        pay_exchange_notionals = dict_init['pay_exchange_notionals']
        pay_MTM_notional_currency = dict_init['pay_MTM_notional_currency']
        # Pay cpn Details
        pay_leg_type = dict_init['pay_leg_type']
        pay_fixed_rate = dict_init['pay_fixed_rate']
        pay_float_spread = dict_init['pay_float_spread']
        pay_cpn_cap = dict_init['pay_cpn_cap']
        pay_cpn_floor = dict_init['pay_cpn_floor']
        pay_day_count_basis = dict_init['pay_day_count_basis']
        # Pay pricing yield curves
        pay_discount_curve = dict_init['pay_discount_curve']
        pay_forward_curve = dict_init['pay_forward_curve']
        # Rec Schedule
        rec_payment_frequency = dict_init['rec_payment_frequency']
        rec_roll_convention = dict_init['rec_roll_convention']
        rec_day_roll = dict_init['rec_day_roll']
        rec_stub = dict_init['rec_stub']
        rec_first_stub = dict_init['rec_first_stub']
        rec_last_stub = dict_init['rec_last_stub']
        rec_first_cpn_end_date = dict_init['rec_first_cpn_end_date']
        rec_last_cpn_start_date = dict_init['rec_last_cpn_start_date']
        rec_payment_type = dict_init['rec_payment_type']
        rec_payment_delay = dict_init['rec_payment_delay']
        rec_fixing_days_ahead = dict_init['rec_fixing_days_ahead']
        rec_currency_holidays = dict_init['rec_currency_holidays']
        rec_city_holidays = dict_init['rec_city_holidays']
        rec_holiday_calendar = dict_init['rec_holiday_calendar']
        # Rec Notional & Currencies
        rec_notional = dict_init['rec_notional']
        rec_transaction_price = dict_init['rec_notional']
        rec_notional_currency = dict_init['rec_notional_currency']
        rec_settlement_currency = dict_init['rec_settlement_currency']
        rec_exchange_notionals = dict_init['rec_exchange_notionals']
        rec_MTM_notional_currency = dict_init['rec_MTM_notional_currency']
        # Rec cpn Details
        rec_leg_type = dict_init['rec_leg_type']
        rec_fixed_rate = dict_init['rec_fixed_rate']
        rec_float_spread = dict_init['rec_float_spread']
        rec_cpn_cap = dict_init['rec_cpn_cap']
        rec_cpn_floor = dict_init['rec_cpn_floor']
        rec_day_count_basis = dict_init['rec_day_count_basis']
        # Rec pricing yield curves
        rec_discount_curve = dict_init['rec_discount_curve']
        rec_forward_curve = dict_init['rec_forward_curve']        
        # Margin period of risk details
        
        self.pay_leg = Leg(
            effective_date=effective_date,
            maturity_date=maturity_date,
            tenor=tenor,
            transaction_date=transaction_date, 
            forward_starting=forward_starting, 
            transaction_price=pay_transaction_price,
            payment_frequency=pay_payment_frequency,
            roll_convention=pay_roll_convention,
            day_roll=pay_day_roll,
            stub=pay_stub,
            first_stub=pay_first_stub,
            last_stub=pay_last_stub,
            first_cpn_end_date=pay_first_cpn_end_date,
            last_cpn_start_date=pay_last_cpn_start_date,
            payment_type=pay_payment_type, 
            payment_delay=pay_payment_delay,
            fixing_days_ahead=pay_fixing_days_ahead,
            currency_holidays=pay_currency_holidays,
            city_holidays=pay_city_holidays,
            holiday_calendar=pay_holiday_calendar,
            notional=pay_notional,
            notional_currency=pay_notional_currency,
            settlement_currency=pay_settlement_currency,
            exchange_notionals=pay_exchange_notionals,
            MTM_notional_currency=pay_MTM_notional_currency,
            pay_rec='pay',
            leg_type=pay_leg_type,
            fixed_rate=pay_fixed_rate,
            float_spread=pay_float_spread,
            cpn_cap=pay_cpn_cap,
            cpn_floor=pay_cpn_floor,
            day_count_basis=pay_day_count_basis,
            dsc_crv=pay_discount_curve,
            fwd_crv=pay_forward_curve)
            
        self.rec_leg = Leg(
            effective_date=effective_date,
            maturity_date=maturity_date,
            tenor=tenor,
            transaction_date=transaction_date, 
            transaction_price=rec_transaction_price,
            forward_starting=forward_starting, 
            payment_frequency=rec_payment_frequency,
            roll_convention=rec_roll_convention,
            day_roll=rec_day_roll,
            stub=rec_stub,
            first_stub=rec_first_stub,
            last_stub=rec_last_stub,
            first_cpn_end_date=rec_first_cpn_end_date,
            last_cpn_start_date=rec_last_cpn_start_date,
            payment_type=rec_payment_type, 
            payment_delay=rec_payment_delay,
            fixing_days_ahead=rec_fixing_days_ahead,
            currency_holidays=rec_currency_holidays,
            city_holidays=rec_city_holidays,
            holiday_calendar=rec_holiday_calendar,
            notional=rec_notional,
            notional_currency=rec_notional_currency,
            settlement_currency=rec_settlement_currency,
            exchange_notionals=rec_exchange_notionals,
            MTM_notional_currency=rec_MTM_notional_currency,
            leg_type=rec_leg_type,
            pay_rec='rec',
            fixed_rate=rec_fixed_rate,
            float_spread=rec_float_spread,
            cpn_cap=rec_cpn_cap,
            cpn_floor=rec_cpn_floor,
            day_count_basis=rec_day_count_basis,
            dsc_crv=rec_discount_curve,
            fwd_crv=rec_forward_curve)
        
    def price(self, 
              calc_PV01=False):
        
        pay_pricing = self.pay_leg.price(calc_PV01=calc_PV01)
        rec_pricing = self.rec_leg.price(calc_PV01=calc_PV01)
        
        pricing = {}
        pricing['price'] = pay_pricing['price'] + rec_pricing['price'] 
        pricing['price_pay_leg'] = pay_pricing['price']
        pricing['price_rec_leg'] = rec_pricing['price']
        pricing['cashflows_pay_leg'] = pay_pricing['cashflows']
        pricing['cashflows_rec_leg'] = rec_pricing['cashflows']
        
        if calc_PV01:            
            dict_PV01 = {'PV01': (pay_pricing['price_discount_and_forward_shift'] + rec_pricing['price_discount_and_forward_shift']) - pricing['price'],
                         'PV01_discount': (pay_pricing['price_discount_shift'] + rec_pricing['price_discount_shift']) - pricing['price'],
                         'PV01_forward': (pay_pricing['price_forward_shift'] + rec_pricing['price_forward_shift']) - pricing['price'],
                         'PV01_pay_leg_discount_shift': pay_pricing['price_discount_shift'] - pay_pricing['price'],
                         'PV01_pay_leg_forward_shift': pay_pricing['price_forward_shift'] - pay_pricing['price'],
                         'PV01_pay_leg_discount_and_forward_shift': pay_pricing['price_discount_and_forward_shift'] - pay_pricing['price'], 
                         'PV01_rec_leg_discount_shift': rec_pricing['price_discount_shift'] - rec_pricing['price'],
                         'PV01_rec_leg_forward_shift': rec_pricing['price_forward_shift'] - rec_pricing['price'],
                         'PV01_rec_leg_discount_and_forward_shift': rec_pricing['price_discount_and_forward_shift'] - rec_pricing['price']} 
            pricing['PV01'] = dict_PV01
            
        return pricing
        
    def solve_to_par(self,
                     leg_to_adjust: str='pay_leg') -> (float, str):
        
        if leg_to_adjust == 'pay_leg':
            pricing_rec_leg = self.rec_leg.price()
            rate, msg = self.pay_leg.solver(solve_price=-pricing_rec_leg['price'])
        elif leg_to_adjust == 'rec_leg':
            pricing_pay_leg = self.pay_leg.price()
            rate, msg = self.rec_leg.solver(solve_price=-pricing_pay_leg['price'])
        
        return rate, msg

            
    def set_pay_leg_discount_curve(self, zero_crv):
        self.pay_leg.dsc_crv = zero_crv
        
    def set_rec_leg_discount_curve(self, zero_crv):   
        self.rec_leg.dsc_crv = zero_crv

    def set_discount_curve(self, zero_crv):   
        self.set_pay_leg_discount_curve(zero_crv)
        self.set_rec_leg_discount_curve(zero_crv)
    
    def set_pay_leg_forward_curve(self, zero_crv):
        self.pay_leg.fwd_crv = zero_crv
        
    def set_rec_leg_forward_curve(self, zero_crv):   
        self.rec_leg.fwd_crv = zero_crv

    def set_forward_curve(self, zero_crv):   
        self.set_pay_leg_forward_curve(zero_crv)
        self.set_rec_leg_forward_curve(zero_crv) 
                
            

                
                
                
                
                
                
                
                
                
        
        
               


# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    import sys
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve()) 
    sys.path.append(os.getcwd())
    print('__main__ - current working directory:', os.getcwd())
   

from frm.pricing_engine.garman_kohlhagen import garman_kohlhagen
import numpy as np
import scipy.fft
import scipy.optimize 
from scipy.integrate import quad
from scipy.optimize import minimize, newton, root_scalar
from numba import jit
from typing import Tuple
import warnings
from scipy.stats import norm

def solve_implied_σ(X: float, S: float, K: float, r_d: float, r_f: float, tau: float, cp: int, σ_guess: float) -> float:
    """
    Solve the implied volatility using the Garman-Kohlhagen model for a European Vanilla FX option.

    Parameters:
    - X (float): Option price
    - S (float): Spot price
    - K (float): Strike price
    - r_d (float): Domestic interest rate
    - r_f (float): Foreign interest rate
    - tau (float): Time to maturity in years
    - cp (int): Option type (1 for call, -1 for put)
    - σ_guess (float): Initial guess for volatility

    Returns:
    - float: Implied volatility

    Attempts to find implied volatility using Newton's method initially and falls back to Brent's method in case of failure.
    """
    try:
        # Try Netwon's method first (it's faster but less robust)
        return newton(lambda σ: (garman_kohlhagen(S=S,σ=σ,r_f=r_f,r_d=r_d,tau=tau,cp=cp,K=K,task='px') - X), x0=σ_guess, tol=1e-4, maxiter=50)
    except RuntimeError:
        # Fallback to Brent's method
        return root_scalar(lambda σ: (garman_kohlhagen(S=S,σ=σ,r_f=r_f,r_d=r_d,tau=tau,cp=cp,K=K,task='px') - X), bracket=[0.0001, 2], method='brentq').root

def heston_fit_vanilla_fx_smile(
        Δ: np.array, 
        Δ_convention: str, 
        σ_market: np.array, 
        S: float, 
        r_f: float, 
        r_d: float, 
        tau: float, 
        cp: np.array,
        pricing_method='carr_madan_gauss_kronrod_quadrature') -> Tuple[float, float, float, float, float, float, np.array, float]:
    """
    Fit the Heston model to the FX market implied volatility smile.
    
    Parameters:
    - Δ (np.array): Vector of spot delta values
    - Δ_convention (str): Delta convention ('prem-adj' or 'prem-adj-fwd')
    - σ_market (np.array): Vector of market implied volatilities
    - S (float): Spot price
    - r_f (float): Foreign interest rate
    - r_d (float): Domestic interest rate
    - tau (float): Time to maturity in years
    - cp (np.array): Vector of option types (1 for call, -1 for put)

    Returns:
    - Tuple: Initial volatility (v0), vol of vol (vv), mean reversion (kappa), long-run mean (theta), market price of volatility risk (lambda_), correlation (rho), vector of implied volatilities (IV), sum of squared errors (SSE)

    References:
    [1] Janek, A., Kluge, T., Weron, R., Wystup, U. (2010). "FX smile in the Heston model"

    The Heston SDE is:
        dS(t) = µ*S*dt + σ(t)*S*dW1(t)
        dσ(t) = kappa(theta - σ(t))*dt + vv*σ(t)*dW2(t)

    The Heston model is defined by six parameters 
    - v0: Initial volatility.
    - vv: Volatility of volatility.
    - kappa: rate of mean reversion to the long-run variance
    - theta: Long-run variance.
    - lambda_: Market price of volatility risk.
    - rho: Correlation.

    The function initially estimates strikes using the Garman-Kohlhagen model. 
    It then optimizes the Heston parameters to minimize the sum of squared errors between market and model-implied volatilities.
    """
    
    def heston_vanilla_sse(param, cp, S, strikes, v0, σ_market, r_d, r_f, tau, kappa):        
        """
        Compute the sum of squared errors (SSE) between market and model implied volatilities.
    
        Parameters:
        param (list): [vol of vol (vv), long-run variance (theta), correlation (rho)]
        cp (array): Option types; 1 for call, -1 for put.
        S (float): Spot price.
        strikes (array): Vector of strike prices.
        v0 (float): Initial volatility.
        σ_market (array): Vector of market implied volatilities.
        r_d (float): Domestic interest rate (annualized).
        r_f (float): Foreign interest rate (annualized).
        tau (float): Time to maturity in years.
        kappa (float): Level of mean reversion.
    
        Returns:
        float: Sum of squared errors between model and market implied volatilities.
        """

        vv, theta, rho = param
        if vv < 0.0 or theta < 0.0 or abs(rho) > 1.0:
            warnings.warn("Invalid value for vv, theta or rho encountered")
            return np.inf
        P = np.zeros(nb_strikes)
        IV = np.zeros(nb_strikes)
        for i in range(nb_strikes):
            if pricing_method == 'analytical_1993':
                P[i] = heston1993_price_fx_vanilla_european(cp[i], S, strikes[i], v0, vv, r_d, r_f, tau, kappa, theta, lambda_, rho)
            elif pricing_method == 'carr_madan_fft_w_simpsons':
                P[i] = heston_carr_madan_fx_vanilla_european(cp[i], S, strikes[i], tau, r_d, r_f, kappa, theta, vv, rho, v0, integration_method=1)
            elif pricing_method == 'carr_madan_gauss_kronrod_quadrature':
                P[i] = heston_carr_madan_fx_vanilla_european(cp[i], S, strikes[i], tau, r_d, r_f, kappa, theta, vv, rho, v0, integration_method=0)
            elif pricing_method == 'lipton gauss-kronrod quadrature':
                pass # not working
                #P[i] = heston_lipton_price_fx_vanilla_european(cp[i], S, strikes[i], tau, r_d, r_f, kappa, theta, vv, rho, v0)
            
            if P[i] < 0:
                print("Reeeee")
                pass
            
            IV[i] = solve_implied_σ(P[i], S, strikes[i], r_d, r_f, tau, cp[i], σ_market[i])
        return np.sum((σ_market - IV)**2)

    # Calculate strikes for market deltas
    strikes = garman_kohlhagen(S=S,σ=σ_market,r_f=r_f, r_d=r_d,tau=tau,Δ=Δ,task='strike',Δ_convention=Δ_convention)
    nb_strikes = len(strikes)
    
    # Set the initial variance, v0, to the implied ATM market volatility
    # v0 will NOT be solved in the calibration
    v0 = (σ_market[nb_strikes // 2])**2 
    
    # Set mean reversion to the long-run variance to constant value = 1.5
    # The influence of mean reversion can be compensated by a the volatility of variance
    # kappa could be adjusted to enforce a valid Feller condition but this may detrimentally impact the fit to the market 
    # The Feller condition is often required to be violated in order to get a good fit to market data
    kappa = 1.5 
    
    # lambda_ is the market price of volatility risk
    # lambda_ will NOT be solved in the calibration
    # We set it to zero, given we are calibrating off option prices which already contain this feature
    # If calibrating of historical data, this field would need to be considered.
    lambda_ = 0 
    
    # Set initial values for vv, theta, rho (the parameters we are solving for)
    initparam = [2 * np.sqrt(v0), 2 * v0, 0]

    res = minimize(lambda param: heston_vanilla_sse(param, cp, S, strikes, v0, σ_market, r_d, r_f, tau, kappa), initparam)
    vv, theta, rho = res.x
    
    if 2 * kappa * theta - vv**2 <= 0.0:
        # In the Heston model, the Feller condition is often required to be violated in order to get a good fit to market data
        warnings.warn("Feller condition violated.") 

    # Calculate the Heston model implied volatilities so we can chart and compare them to σ_market
    P = np.zeros(nb_strikes)
    IV = np.zeros(nb_strikes)
    for i in range(nb_strikes):
        P[i] = heston1993_price_fx_vanilla_european(cp[i], S, strikes[i], v0, vv, r_d, r_f, tau, kappa, theta, lambda_, rho)
        IV[i] = solve_implied_σ(P[i], S, strikes[i], r_d, r_f, tau, cp[i], σ_market[i])
 
    SSE = heston_vanilla_sse(res.x, cp, S, strikes, v0, σ_market, r_d, r_f, tau, kappa)
    
    if SSE == np.inf:
        # Calibration failed
        return None
    else:    
        return v0, vv, kappa, theta, lambda_, rho, IV, SSE

#%% Heston 1993 analytical pricing implementation (2nd version of the Characteristic function)

@jit(nopython=True)
def heston_1993_fx_vanilla_european_integral(phi, m, S, k, v0, vv, r_d, r_f, tau, kappa, theta, lambda_, rho):
    """
    Defines the integral for pricing an FX Vanilla European option per the analytic Heston 1993 formula
    This implementation uses the 2nd form of the Heston Characteristic function, detailed in Albrecher 2006 is used as it is more numerically stable.
    This in an auxiliary function for heston1993_price_fx_vanilla_european(), separated so @jit can be used
    
    Parameters:
    phi (complex): Point at which the auxiliary function is evaluated.
    m (int): Index for specific calculation (either 1 or 2).
    s (float): Spot price of the underlying.
    k (float): Strike price.
    v0 (float): Initial volatility.
    vv (float): Volatility of volatility.
    r_d (float): Domestic interest rate.
    r_f (float): Foreign interest rate.
    tau (float): Time to maturity in years.
    kappa (float): Mean reversion level.
    theta (float): Long-term variance.
    lambda_ (float): Market price of volatility risk.
    rho (float): Correlation.

    Returns:
    float: Value of the auxiliary function at point `phi`.
    
    References:
    [1] S.Heston, (1993) A Closed-Form Solution for Options with Stochastic Volatility with Applications to Bond and Currency Options
    [2] H.Albrecher, P.Mayer, W.Schoutens, J.Tistaert (2006) The little Heston trap, Wilmott Magazine, January: 83–92.
    [3] A.Janek, T.Kluge, R.Weron, U.Wystup (2010) FX smile in the Heston model.
    """
    
    # x per equation (11) from Heston, 1993
    x = np.log(S) 
    # a, u, b per equation (12) from Heston, 1993
    a = kappa * theta 
    u = [0.5, -0.5] 
    b = [kappa + lambda_ - rho * vv, kappa + lambda_] 
    
    # d is per per equation (17) from Heston, 1993
    d = np.sqrt((1j * rho * vv * phi - b[m-1]) ** 2 - vv ** 2 * (2 * u[m-1] * phi * 1j - phi ** 2))
    
    # There are two formula's for "g", in the Heston characteristic function because d, an input to g, has two roots. 
    # Let "g1", be the presentation in the original Heston 1993 paper.
    # Let "g2", be the alternative, that is detailed in Albrecher 2006, "The Little Heston Trap". 
    # g2 = 1 / g1. 
    # We use g2, as using g1 leads to numerical problems based on most software implementations of calculation involving complex numbers
    g2 = (b[m-1] - rho * vv * phi * 1j - d) \
        / (b[m-1] - rho * vv * phi * 1j + d)

    D = (b[m-1] - rho * vv * phi * 1j - d) / (vv ** 2) * ((1 - np.exp(-d * tau)) / (1 - g2 * np.exp(-d * tau)))
    C = (r_d - r_f) * phi * 1j * tau + a / (vv ** 2) * ((b[m-1] - rho * vv * phi * 1j - d) * tau - 2 * np.log((1 - g2 * np.exp(-d * tau)) / (1 - g2)))
   
    # f per equation (17) from Heston, 1993
    f = np.exp(C + D * v0 + 1j * phi * x)
    
    # Function inside the integral in equation (18) from Heston, 1993
    F = np.real(np.exp(-1j * phi * np.log(k)) * f / (1j * phi))
    return F


def heston1993_price_fx_vanilla_european(cp, s, k, v0, vv, r_d, r_f, tau, kappa, theta, lambda_, rho):
    """
    Calculate the price of a European Vanilla FX option using the analytical Heston 1993 formulae
    The 2nd form of the Heston Characteristic function, detailed in Albrecher 2006 is used as it is more numerically stable.
    
    Parameters:
    cp (int): Call (1) or Put (-1) option.
    s (float): Spot price.
    k (float): Strike price.
    v0 (float): Initial volatility.
    vv (float): Volatility of volatility.
    r_d (float): Domestic interest rate.
    r_f (float): Foreign interest rate.
    tau (float): Time to maturity in years.
    kappa (float): Level of mean reversion to the long-run variance
    theta (float): Long-run variance.
    lambda_ (float): Market price of volatility risk.
    rho (float): Correlation.

    Returns:
    float: Option price.

    Example:
    >>> heston_garman_kohlhagen(1, 1.03, 1, 0.01, 0.02, 0.05, 0.03, 0.25, 10, 0.01, 0, 0.5)

    References:
    [1] S.Heston, (1993) A Closed-Form Solution for Options with Stochastic Volatility with Applications to Bond and Currency Options
    [2] H.Albrecher, P.Mayer, W.Schoutens, J.Tistaert (2006) The little Heston trap, Wilmott Magazine, January: 83–92.
    [3] A.Janek, T.Kluge, R.Weron, U.Wystup (2010) FX smile in the Heston model.
    
    Written by Agnieszka Janek and Rafal Weron (2010.07.07)
    Revised by Rafal Weron (2010.10.08, 2010.12.27)
    """

    # Equation (18) from Heston, 1993
    P1 = 0.5 + 1/np.pi * quad(lambda phi: heston_1993_fx_vanilla_european_integral(phi, 1, s, k, v0, vv, r_d, r_f, tau, kappa, theta, lambda_, rho), 0, np.inf, epsrel=1e-8)[0]
    P2 = 0.5 + 1/np.pi * quad(lambda phi: heston_1993_fx_vanilla_european_integral(phi, 2, s, k, v0, vv, r_d, r_f, tau, kappa, theta, lambda_, rho), 0, np.inf, epsrel=1e-8)[0]

    Pplus = (1 - cp) / 2 + cp * P1   # Pplus = N(d1)
    Pminus = (1 - cp) / 2 + cp * P2  # Pminus = N(d2)

    # tk is this for spot or forward Δ, matlab code assumed spot delta input?
    X = cp * (s * np.exp(-r_f * tau) * Pplus - k * np.exp(-r_d * tau) * Pminus) 
    
    return X

#%%  Heston model using Carr-Madan via 
# (method=0) Gauss-Kronrod quadrature
# (method=1) Fast Fourier Transform + Simpson method 

def heston_carr_madan_fx_vanilla_european(cp, S, K, T, r_d, r_f, kappa, theta, vv, rho, v0, integration_method=0):
    """
     Calculate European FX option price using the Heston model via Carr-Madan approach.
     
     Parameters:
     cp (int): 1 for call, -1 for put.
     S, K, T (float): Spot price, strike, and time to maturity.
     r_d, r_f (float): Domestic and foreign interest rates.
     kappa, theta, vv, rho, v0 (float): Heston parameters.
     integration_method (int, optional): 0 for Gauss-Kronrod quadrature, 1 for FFT + Simpson's rule. Default is 0.
     
     Returns:
     float: Option price.
     
     References:
     [1] Albrecher et al. (2006) "The little Heston trap."
     [2] Carr, Madan (1998) "Option valuation using the Fast Fourier transform."
     [3] Janek et al. (2010) "FX smile in the Heston model."
     [4] Schmelzle (2010) "Option Pricing Formulae using Fourier Transform."
     
     Authors:
     Written by Agnieszka Janek (2010.07.23)
     Revised by Rafal Weron (2010.10.08)
     Revised by Agnieszka Janek and Rafal Weron (2010.10.21, 2010.12.27)
    """    
    
    if cp == 1:
        alpha = 0.75
    elif cp == -1:
        alpha = 1.75
    
    s0 = np.log(S)
    k = np.log(K)
    
    if integration_method == 0:
        # Integrate using adaptive Gauss-Kronrod quadrature
        result, _ = quad(heston_fft_fx_vanilla_european_integral, 0, np.inf, args=(cp, s0, k, T, r_d, r_f, kappa, theta, vv, rho, v0, alpha))
        y = np.exp(-cp * k * alpha) * result / np.pi
    elif integration_method == 1:
        # Fast Fourier Transform with Simpson's rule (as suggested in [2])
        N = 2**10
        eta = 0.25
        v = np.arange(0, N) * eta
        
        lambda_ = 2*np.pi/(N*eta) 
        b = N*lambda_/2 # Equation (20) per Carr-Madan, 1998
        ku = -b + lambda_ * np.array(range(0,N)) # Equation (19) per Carr-Madan, 1998
        
        u = v - (cp * alpha + 1) * 1j
        d, g, A, B, C, charFunc = char_func(u, s0, k, T, r_d, r_f, kappa, theta, vv, rho, v0)
        F = charFunc * np.exp(-r_d * T) / (alpha**2 + cp * alpha - v**2 + 1j * (cp * 2 * alpha + 1) * v)
        
        # Use Simpson's approximation to calculate FFT (see [2])
        simpson_weights = get_simpson_weights(N) 
        fft_func = np.exp(1j * b * v) * F * eta * simpson_weights
        payoff = np.real(scipy.fft.fft(fft_func))
        option_value = np.exp(-cp * ku * alpha) * payoff / np.pi
        
        y = np.interp(k, ku, option_value)
                
    return y

def get_simpson_weights(n):
    # 1/3, then alternating 2/3, 4/3, 2/3, 4/3, 2/3, 4/3, .... 
    weights = np.array([1] + [4, 2] * ((n-2) // 2) + [1])
    return weights / 3

@jit(nopython=True)
def heston_fft_fx_vanilla_european_integral(v, cp, s0, k, T, r_d, r_f, kappa, theta, vv, rho, v0, alpha):
    """
    Auxiliary function for HESTONFFTVANILLA.

    Parameters:
        v: Evaluation points for auxiliary function.
        cp: Option type (1 for call, -1 for put).
        s0: Log of spot price.
        k: Log of strike price.
        T: Time to maturity (years).
        r: Domestic interest rate.
        rf: Foreign interest rate.
        kappa: Level of mean reversion.
        theta: Long-run variance.
        vv: Volatility of volatility.
        rho: Correlation coefficient.
        v0: Initial volatility.
        alpha: Damping coefficient.

    Returns:
        The values of the auxiliary function evaluated at points v.
    """
    
    u = v - (cp * alpha + 1) * 1j
    d, g, A, B, C, charFunc = char_func(u, s0, k, T, r_d, r_f, kappa, theta, vv, rho, v0)
    ftt_func = charFunc * np.exp(-r_d * T) / (alpha**2 + cp * alpha - v ** 2 + 1j * (cp * 2 * alpha + 1) * v)
    return np.real(np.exp(-1j * v * k) * ftt_func)

@jit
def char_func(u, s0, k, tau, r_d, r_f, kappa, theta, vv, rho, v0):
    """
    Compute the characteristic function for the Heston model.
    
    Parameters:
    u (float or np.array): The argument of the characteristic function.
    s0 (float): Initial stock price.
    k (float): Strike price.
    tau (float): Time to maturity.
    r_d (float): Domestic risk-free interest rate.
    r_f (float): Foreign risk-free interest rate.
    kappa (float): Rate of mean reversion.
    theta (float): Long-term level of volatility.
    vv (float): Volatility of volatility.
    rho (float): Correlation between stock price and volatility.
    v0 (float): Initial volatility level.
    
    Returns:
    tuple: Tuple containing:
        d (float or np.array): A derived value used for further calculations.
        g (float or np.array): A derived value used for further calculations.
        A (float or np.array): A term in the characteristic function.
        B (float or np.array): A term in the characteristic function.
        C (float or np.array): A term in the characteristic function.
        charFunc (float or np.array): Value of the characteristic function at u.
    """    
    
    # Characteristic function (see [1])
    d = np.sqrt((rho * vv * u * 1j - kappa) ** 2 + vv ** 2 * (1j * u + u ** 2))
    g = (kappa - rho * vv * 1j * u - d) / (kappa - rho * vv * 1j * u + d)
    A = 1j * u * (s0 + (r_d - r_f) * tau)
    B = theta * kappa * vv ** -2 * ((kappa - rho * vv * 1j * u - d) * tau - 2 * np.log((1 - g * np.exp(-d * tau)) / (1 - g)))
    C = v0 * vv ** -2 * (kappa - rho * vv * 1j * u - d) * (1 - np.exp(-d * tau)) / (1 - g * np.exp(-d * tau))
    charFunc = np.exp(A + B + C)
    return d, g, A, B, C, charFunc



#%%

def call_put_option_price_cos_method(cf, cp, S0, r, tau, K, N, L):
    """
    Computes the call or put option prices using the COS method.
    
    Parameters:
    - cf (function): Characteristic function
    - cp (int): 1 for call and -1 for put
    - S0 (float): Initial stock price
    - r (float): Interest rate
    - tau (float): Time to maturity
    - K (list or np.array): List of strike prices
    - N (int): Number of expansion terms
    - L (float): Size of truncation domain
    
    Returns:
    - np.array: Option prices
    """    
    
    # cf - Characteristic function as a functon, in the book denoted by phi
    # cp - C for call and P for put
    # S0 - Initial stock price
    # r - Interest rate (constant)
    # tau - Time to maturity
    # K - List of strikes
    # N - Number of expansion terms
    # L - Size of truncation domain (typ.:L=8 or L=10)

    # Reshape K to become a column vector
    if K is not np.array:
        K = np.array(K).reshape([len(K),1])

    x0 = np.log(S0 / np.array(K)).reshape(-1, 1)

    # Truncation domain
    a, b = -L * np.sqrt(tau), L * np.sqrt(tau)
    
    # Summation from k=0 to k=N-1
    k = np.linspace(0, N-1, N).reshape(-1, 1)
    u = k * np.pi / (b - a)
    
    H_k = call_put_coefficients(cp, a, b, k)

    mat = np.exp(1j * np.outer((x0 - a) , u))
    temp = cf(u) * H_k
    temp[0] = 0.5 * temp[0]
    value = np.exp(-r * tau) * K * np.real(mat.dot(temp))

    return value


def call_put_coefficients(cp, a, b, k):
    """
    Compute coefficients for call or put options using COS method.
    
    Parameters:
    - cp (int): 1 for call, -1 for put
    - a (float): Lower truncation boundary
    - b (float): Upper truncation boundary
    - k (np.array): Array of k values
    
    Returns:
    - np.array: Coefficients
    """        
    c, d = (0, b) if cp == 1 else (a, 0)
    coef = chi_psi(a, b, c, d, k)
    
    chi, psi = coef['chi'], coef['psi']
    prefactor = 2 / (b - a)
    
    return prefactor * (chi - psi) if cp == 1 else prefactor * (-chi + psi)


@jit(nopython=True)
def chi_psi(a, b, c, d, k):
    """
    Compute chi and psi coefficients for COS method.
    
    Parameters:
    - a, b, c, d (float): Boundaries for truncation and integration
    - k (np.array): Array of k values
    
    Returns:
    - dict: {'chi': np.array, 'psi': np.array}
    """        
    psi = np.sin(k * np.pi * (d - a) / (b - a)) - np.sin(k * np.pi * (c - a)/ (b - a))
    psi[1:] = psi[1:] * (b - a) / (k[1:] * np.pi)
    psi[0] = d - c
    
    chi = 1.0 / (1.0 + np.power((k * np.pi / (b - a)) , 2.0))
    expr1 = np.cos(k * np.pi * (d - a)/(b - a)) * np.exp(d) - np.cos(k * np.pi  * (c - a) / (b - a)) * np.exp(c)
    expr2 = k * np.pi / (b - a) * np.sin(k * np.pi * (d - a) / (b - a)) - k * np.pi / (b - a) * np.sin(k * np.pi * (c - a) / (b - a)) * np.exp(c)
    chi = chi * (expr1 + expr2)
    
    return {"chi": chi, "psi": psi}


def bs_call_option_price(cp, S_0, K, σ, tau, r):
    """
    Computes Black-Scholes call or put option prices.
    
    Parameters:
    - cp (int): 1 for call, -1 for put
    - S_0 (float): Initial stock price
    - K (list or np.array): List of strike prices
    - σ (float): Volatility
    - tau (float): Time to maturity
    - r (float): Interest rate
    
    Returns:
    - np.array: Option prices
    """        
    K = np.array(K).reshape(-1, 1)
    d1 = (np.log(S_0 / K) + (r + 0.5 * σ ** 2) * tau) / (σ * np.sqrt(tau))
    d2 = d1 - σ * np.sqrt(tau)

    if cp == 1:
        return norm.cdf(d1) * S_0 - norm.cdf(d2) * K * np.exp(-r * tau)
    elif cp == -1:
        return norm.cdf(-d2) * K * np.exp(-r * tau) - norm.cdf(-d1) * S_0


# Implied volatility method
def implied_volatility(cp, market_price, K, T, S_0, r):
    """
    Calculates implied volatility using Newton's method.
    
    Parameters:
    - cp (int): 1 for call, -1 for put
    - market_price (float): Observed market price of option
    - K (float): Strike price
    - T (float): Time to maturity
    - S_0 (float): Initial stock price
    - r (float): Interest rate
    
    Returns:
    - float: Implied volatility
    """        
    func = lambda sigma: (bs_call_option_price(cp, S_0, K, sigma, T, r) - market_price) ** 1.0
    return scipy.optimize.newton(func, 0.7, tol=1e-5)


def chf_heston_model(r, tau, kappa, gamma, vbar, v0, rho):
    """
    Characteristic function for Heston model.
    
    Parameters:
    - r (float): Interest rate
    - tau (float): Time to maturity
    - kappa (float): Mean reversion rate
    - gamma (float): Volatility of volatility
    - vbar (float): Long-term average volatility
    - v0 (float): Initial variance
    - rho (float): Correlation coefficient
    
    Returns:
    - function: Characteristic function for Heston model
    """        
    D1 = lambda u: np.sqrt((kappa - gamma * rho * 1j * u) ** 2 + (u ** 2 + 1j * u) * gamma ** 2)
    g = lambda u: (kappa - gamma * rho * 1j * u - D1(u)) / (kappa - gamma * rho * 1j * u + D1(u))
    C = lambda u: (1 - np.exp(-D1(u) * tau)) / (gamma ** 2 * (1 - g(u) * np.exp(-D1(u) * tau)))
    A = lambda u: r * 1j * u * tau + kappa * vbar * tau / gamma ** 2 * (kappa - gamma * rho * 1j * u - D1(u)) \
                 - 2 * kappa * vbar / gamma ** 2 * np.log((1 - g(u) * np.exp(-D1(u) * tau)) / (1 - g(u)))
    return lambda u: np.exp(A(u) + C(u) * v0)



#%%  Heston model using the Lewis-Lipton formula (something wrong with it)

# def heston_lipton_fx_vanilla_european_integral(v, S, K, T, r_d, r_f, kappa, theta, vv, rho, v0):
#     """
#     Auxiliary function used by heston_vanilla_lipton for numerical integration.
#     This in an auxiliary function for heston1993_price_fx_vanilla_european(), separated so @jit can be used
    
#     Parameters:
#     v (float): Variable for integration.
#     S (float): Spot price.
#     K (float): Strike price.
#     T (float): Time to maturity.
#     r (float): Domestic interest rate.
#     rf (float): Foreign interest rate.
#     kappa (float): Rate of mean reversion.
#     theta (float): Long-term volatility level.
#     vv (float): Volatility of volatility.
#     rho (float): Correlation between Wiener processes.
#     v0 (float): Initial volatility.
    
#     Returns:
#     float: Value of the auxiliary function at point v.
#     """        
#     X = np.log(S / K) + (r_d - r_f) * T
#     kappa_hat = kappa - rho * vv / 2
#     zeta = np.sqrt(v ** 2 * vv ** 2 * (1 - rho ** 2) + 2 * 1j * v * vv * rho * kappa_hat + kappa_hat ** 2 + vv ** 2 / 4)
#     psi_plus = -(1j * kappa * rho * vv + kappa_hat) + zeta
#     psi_minus = (1j * kappa * rho * vv + kappa_hat) + zeta
#     alpha = -kappa * theta / (vv ** 2) * (psi_plus * T + 2 * np.log((psi_minus + psi_plus * np.exp(-zeta * T)) / (2 * zeta)))
#     beta = (1 - np.exp(-zeta * T)) / (psi_minus + psi_plus * np.exp(-zeta * T))
#     payoff = np.real(np.exp((-1j * v + 0.5) * X + alpha - (v ** 2 + 0.25) * beta * v0) / (v ** 2 + 0.25))
#     return payoff

# def heston_lipton_price_fx_vanilla_european(cp, S, K, T, r_d, r_f, kappa, theta, sigma, rho, v0):
#     """
#     Calculate European FX option price using the Heston model and Lewis-Lipton formula.
    
#     Parameters:
#     phi (int): 1 for call option, -1 for put option.
#     S (float): Spot price.
#     K (float): Strike price.
#     T (float): Time to maturity.
#     r_d (float): Domestic interest rate.
#     r_f (float): Foreign interest rate.
#     kappa (float): Rate of mean reversion.
#     theta (float): Long-term volatility level.
#     sigma (float): Volatility of volatility.
#     rho (float): Correlation between Wiener processes.
#     v0 (float): Initial volatility.
    
#     Returns:
#     float: Option price.
#     """    
#     C = np.exp(-r_f * T) * S - np.exp(-r_d * T) * K / np.pi * quad(heston_lipton_fx_vanilla_european_integral, 0, np.inf, args=(S, K, T, r_d, r_f, kappa, theta, sigma, rho, v0), epsrel=1e-8)[0]
#     if cp == 1:  # call option
#         X = C
#     else:  # put option
#         X = C - S * np.exp(-r_f * T) + K * np.exp(-r_d * T)
#     return X

#%%





# -*- coding: utf-8 -*-


import os
import pathlib

os.chdir(pathlib.Path(__file__).parent.parent.resolve()) 

import numpy as np
import multiprocessing as mp
import time

NUMBER_OF_CORES = mp.cpu_count()
APPLY_CONTROL_VARIATES = False
APPLY_PARALLEL_PROCESSING = True

np.random.seed(1) # Set random seed so we get repeatable results

def GeometricBrownian(x0, 
                      dt, 
                      drift, 
                      speed, 
                      W, 
                      correlation_matrix=None):
    """
    Generate the path of a Geomtric Brownian motion.

        X(t) = X(0) + N(drift * t, speed**2 * t; 0, t)

    where N(a,b; t0, t1) is a normally distributed random variable with mean a and
    variance b.  The parameters t0 and t1 make explicit the statistical
    independence of N on different time intervals; that is, if [t0, t1) and
    [t2, t3) are disjoint intervals, then N(a, b; t0, t1) and N(a, b; t2, t3)
    are independent.
    
    The random variable of the position at time t, X(t), has a normal distribution whose mean is
    the position at time t=0 + drift*t and whose variance is speed**2*t.
    
    Written as an iteration scheme,

        X(t + dt) = X(t) * exp(N(drift * dt, speed**2 * dt)) 

    Arguments
    ---------
    x0 : float
        The initial condition(s) (i.e. position(s)) of the Brownian motion.
    dt : float
        The time step.
    drift : float
        Specifies the "drift" of the Brownian motion. 
    speed : float
        Specifies "speed" of the Brownian motion.  
    W : numpy array  
        Random Normal Variates

    Returns
    -------
    A numpy array of floats with shape `x0.shape + (n,)`.
    
    Note that the initial value `x0` is not included in the returned array.
    """
    
    nb_timesteps = 365
    nb_assets = 1
    nb_iterations = 100000 
    
    
    def helper(x0, dt, drift, speed, nb_timesteps, nb_assets, nb_iterations):
        np.random.seed()
        
        sqrt_dt = np.sqrt(dt)
        
        if APPLY_CONTROL_VARIATES:
            # https://en.wikipedia.org/wiki/Control_variates
            W_1 = np.random.randn(nb_timesteps,nb_assets,int(nb_iterations / 2))
            result_1 = x0 * np.cumprod(np.exp((drift - 0.5 * speed**2) * dt + speed * sqrt_dt * W_1),axis=1)  
            W_2 = -W_1 
            result_2 = x0 * np.cumprod(np.exp((drift - 0.5 * speed**2) * dt + speed * sqrt_dt * W_2),axis=1)
            result = np.append(result_1, result_2, axis=2)
        else:
            W = np.random.randn(nb_timesteps,nb_assets,nb_iterations)
            result = x0 * np.cumprod(np.exp((drift - 0.5 * speed**2) * dt + speed * sqrt_dt * W),axis=1)
            
        return result
            
    if APPLY_PARALLEL_PROCESSING:
        result_list = []
        def log_result(result):
            # This is called whenever helper(i) returns a result.
            # result_list is modified only by the main process, not the pool workers.
            result_list.append(result)
        
        pool = mp.Pool(NUMBER_OF_CORES)
        for i in range(10):
            res = pool.apply_async(helper, args=(x0, dt, drift, speed, nb_timesteps, nb_assets, nb_iterations,), callback=log_result)
            print(res.get())
        pool.close()
        pool.join()
        result = result_list
    else:
        result = helper(x0, dt, drift, speed, nb_timesteps, nb_assets, nb_iterations)


    # https://en.wikipedia.org/wiki/Antithetic_variates

    return result


# Multi core
# https://agustinus.kristia.de/techblog/2016/06/13/parallel-monte-carlo/
#
#


if __name__ == '__main__':

    x0 = 10
    dt = 1/365
    drift = 0.05
    speed = 0.3
    
    
    t1 = time.time()
    
    result = GeometricBrownian(x0, dt, drift, speed, None)
    
    t2 = time.time()
    
    print(t2-t1)


    
    
    


# -*- coding: utf-8 -*-

## Title:        Option Pricing
## Author:       Elisa FLeissner, Lars Stauffenegger
## Email:        elisa.fleissner@student.unisg.ch,
##               lars.stauffenegger@student.unisg.ch,
## Place, Time:  Zürich, 24.03.19
## Description:  Option Pricing with Black Scholes, COS Method and Heston Model
## Improvements: -
## Last changes: -


# In[1]: Packages
import quandl
import numpy as np
import AllFunctions as func
import matplotlib.pyplot as plt


# In[2]: Data
# Import from quandl
quandl.ApiConfig.api_key = "mrMTRoAdPycJSyzyjxPN"
ticker     = "AAPL"
database   = "EOD"
identifier = database + "/" + ticker
stockData  = quandl.get(identifier, rows = 500)

# Return and Volatility
logReturn = np.log(stockData.Close) - np.log(stockData.Close.shift(1))
logReturn.drop(logReturn.index[:1], inplace = True)
tradingDaysCount   = 252
annualisedMean     = np.mean(logReturn) * tradingDaysCount
annualisedVariance = np.var(logReturn) * tradingDaysCount
annualisedStdDev   = np.sqrt(annualisedVariance)
lastPrice          = stockData.Close.tail(1)


# In[2]: Parameter
# Volvol and rho according to Fang, 2010, p. 30
r      = 0                  # assumption Risk-free rate
mu     = r #annualisedMean  # Mean rate of drift
sigma  = annualisedStdDev   # Initial Vola of underyling at time 0; also called u0 or a
S0     = lastPrice[0]       # Today's stock price
tau    = 30 / 365           # Time to expiry in years
q      = 0                  # Divindend Yield
lm     = 1.5768             # The speed of mean reversion
v_bar  = annualisedVariance # Mean level of variance of the underlying
volvol =  0.5751            # Volatility of the volatiltiy process
rho    = -0.5711            # Covariance between the log stock and the variance process

# Range of Strikes
mini    = int(S0 * 0.8)
maxi    = int(S0 * 1.2)
K       = np.arange(mini, maxi, dtype = np.float)

# Truncation Range
L       = 120
a, b    = func.truncationRange(L, mu, tau, sigma, v_bar, lm, rho, volvol)
bma     = b-a

# Number of Points
N       = 15
k       = np.arange(np.power(2,N))

# Input for the Characterstic Function Phi
u       = k * np.pi/bma


# In[3]: Black Scholes Option Pricing
C_BS, P_BS = func.blackScholes(S0, K, r, tau, sigma, q)
print(C_BS)


# In[4]: COS-FFT Value Function for Put
UkPut  = 2 / bma * ( func.cosSer1(a,b,a,0,k) - func.cosSerExp(a,b,a,0,k) )
UkCall = 2 / bma * ( func.cosSerExp(a,b,0,b,k) - func.cosSer1(a,b,0,b,k) )


# In[5]: COS with BS-Characterstic Function
charactersticFunctionBS = func.charFuncBSM(u, mu, sigma, tau)

C_COS = np.zeros((np.size(K)))

for m in range(0,np.size(K)):
    x  = np.log(S0/K[m])
    addIntegratedTerm = np.exp(1j * k * np.pi * (x-a)/bma)
    Fk = np.real(np.multiply(charactersticFunctionBS, addIntegratedTerm))
    Fk[0]=0.5 * Fk[0] 
    C_COS[m] = K[m] * np.sum(np.multiply(Fk,UkCall)) * np.exp(-r * tau)
    
print (C_COS)


# In[6]: COS with Fang & Oosterlee (2008) Version of Heston's Characteristic Function
charactersticFunctionHFO = func.charFuncHestonFO(mu, r, u, tau, sigma, v_bar, lm, rho, volvol)

C_COS_HFO = np.zeros((np.size(K)))
P_COS_HFO = np.zeros((np.size(K)))
C_COS_PCP = np.zeros((np.size(K)))

for m in range(0, np.size(K)):
    x  = np.log(S0/K[m])
    addIntegratedTerm = np.exp(1j * k * np.pi * (x-a)/bma)
    Fk = np.real(charactersticFunctionHFO * addIntegratedTerm)
    Fk[0] = 0.5 * Fk[0]						
    C_COS_HFO[m] = K[m] * np.sum(np.multiply(Fk, UkCall)) * np.exp(-r * tau)
    P_COS_HFO[m] = K[m] * np.sum(np.multiply(Fk, UkPut)) * np.exp(-r * tau)
    C_COS_PCP[m] = P_COS_HFO[m] + S0 * np.exp(-q * tau) - K[m] * np.exp(-r * tau)

print(C_COS_HFO)
print(P_COS_HFO)
print(C_COS_PCP)


# In[7]: Plotting
plt.plot(K, C_BS, "g.", K, C_COS, "b.", K, C_COS_HFO, "r.")
plt.axvline(x = S0)
plt.show()
print("C_BS = green, C_COS = blue, C_COS_HFO = red")

## End

#%%

import numpy as np

cp = 1
S0 = 100
r = 0.00
tau = 1

K = np.linspace(50,150)
K = np.array(K).reshape([len(K),1])

# COS method settings
L = 3

kappa = 1.5768
gamma = 0.5751
vbar = 0.0398
rho =-0.5711
v0 = 0.0175


#%%











#%%
"""
Created on Thu Nov 30 2018
Merton Model and convergence obtained with the COS method
@author: Lech A. Grzelak
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
import enum
import scipy.optimize as optimize
from numba import jit


v0 = 0.0175 # intial variance
vv = 0.5751 # volatility of volatility 
kappa = 1.5768 # rate of mean reversion towards the long-term mean of variance process
theta = 0.0398 # long-term mean of the variance process
rho =-0.5711 # correlation


def chf_heston_model(r, tau, v0, vv, kappa, theta, rho):
    """
    Characteristic function for the Heston model.
    
    Parameters:
    - r (float): Risk-free interest rate
    - tau (float): Time to maturity
    - v0 (float): Initial volatility
    - vv (float): Volatility of volatility
    - kappa (float): rate of mean reversion towards the long-term mean of variance process
    - theta (float): long-term mean of the variance process
    - rho (float): Correlation between price and volatility
    
    Returns:
    - function: Characteristic function taking φ (complex) as input and returning complex value
    """    
    
    D1 = lambda φ: np.sqrt((kappa - vv * rho * 1j * φ) ** 2 + (φ ** 2 + 1j * φ) * vv ** 2)
    g = lambda φ: (kappa - vv * rho * 1j * φ - D1(φ)) / (kappa - vv * rho * 1j * φ + D1(φ))
    C = lambda φ: (1 - np.exp(-D1(φ) * tau)) / (vv ** 2 * (1 - g(φ) * np.exp(-D1(φ) * tau)))
    A = lambda φ: r * 1j * φ * tau + kappa * theta * tau / vv ** 2 * (kappa - vv * rho * 1j * φ - D1(φ)) \
                 - 2 * kappa * theta / vv ** 2 * np.log((1 - g(φ) * np.exp(-D1(φ) * tau)) / (1 - g(φ)))
    return lambda φ: np.exp(A(φ) + C(φ) * v0)


def heston_cos_vanilla_european(cp, S0, r, tau, K, N, L, v0, vv, kappa, theta, rho):
    """
    Computes the call or put option prices using the COS method.
    
    Parameters:
    - cp (int): 1 for call and -1 for put
    - S0 (float): Initial stock price
    - r (float): Interest rate
    - tau (float): Time to maturity
    - K (list or np.array): List of strike prices
    - N (int): Number of expansion terms
    - L (float): Size of truncation domain
    - v0 (float): Initial volatility
    - vv (float): Volatility of volatility
    - kappa (float): rate of mean reversion towards the long-term mean of variance process
    - theta (float): long-term mean of the variance process
    - rho (float): Correlation between price and volatility    
    
    Returns:
    - np.array: Option prices
    """    

    def chi_psi(a, b, c, d, k):
        """
        Compute chi and psi coefficients for COS method.
        
        Parameters:
        - a, b, c, d (float): Boundaries for truncation and integration
        - k (np.array): Array of k values
        
        Returns:
        - tuple (np.array, np.array): chi, psi
        """        
        psi = np.sin(k * np.pi * (d - a) / (b - a)) - np.sin(k * np.pi * (c - a)/ (b - a))
        psi[1:] = psi[1:] * (b - a) / (k[1:] * np.pi)
        psi[0] = d - c
        
        chi = 1.0 / (1.0 + np.power((k * np.pi / (b - a)) , 2.0))
        expr1 = np.cos(k * np.pi * (d - a)/(b - a)) * np.exp(d) - np.cos(k * np.pi  * (c - a) / (b - a)) * np.exp(c)
        expr2 = k * np.pi / (b - a) * np.sin(k * np.pi * (d - a) / (b - a)) - k * np.pi / (b - a) * np.sin(k * np.pi * (c - a) / (b - a)) * np.exp(c)
        chi = chi * (expr1 + expr2)
        
        return chi, psi 

    def call_put_coefficients(cp, a, b, k):
        """
        Compute coefficients for call or put options using COS method.
        
        Parameters:
        - cp (int): 1 for call, -1 for put
        - a (float): Lower truncation boundary
        - b (float): Upper truncation boundary
        - k (np.array): Array of k values
        
        Returns:
        - np.array: Coefficients
        """        
        c, d = (0, b) if cp == 1 else (a, 0)
        chi, psi = chi_psi(a, b, c, d, k)
        prefactor = 2 / (b - a)
        return prefactor * (chi - psi) if cp == 1 else prefactor * (-chi + psi)

    if K is not np.array:
        # Reshape K to become a column vector
        K = np.array(K).reshape([len(K),1]) 
        
    x0 = np.log(S0 / np.array(K)).reshape(-1, 1)

    # Truncation domain
    a, b = -L * np.sqrt(tau), L * np.sqrt(tau)
    
    # Summation from k = 0 to k = N-1
    k = np.linspace(0, N-1, N).reshape(-1, 1)
    u = k * np.pi / (b - a)
    
    H_k = call_put_coefficients(cp, a, b, k)

    mat = np.exp(1j * np.outer((x0 - a) , u))
    cf = chf_heston_model(r, tau, v0, vv, kappa, theta, rho)
    temp = cf(u) * H_k
    temp[0] = 0.5 * temp[0]
    return np.exp(-r * tau) * K * np.real(mat.dot(temp))



def bs_call_option_price(cp, S_0, K, σ, tau, r):
    K = np.array(K).reshape(-1, 1)
    d1 = (np.log(S_0 / K) + (r + 0.5 * σ ** 2) * tau) / (σ * np.sqrt(tau))
    d2 = d1 - σ * np.sqrt(tau)

    if cp == 1:
        return norm.cdf(d1) * S_0 - norm.cdf(d2) * K * np.exp(-r * tau)
    elif cp == -1:
        return norm.cdf(-d2) * K * np.exp(-r * tau) - norm.cdf(-d1) * S_0


# Implied volatility method
def implied_volatility(cp, market_price, K, T, S_0, r):
    func = lambda sigma: (bs_call_option_price(cp, S_0, K, sigma, T, r) - market_price) ** 1.0
    return optimize.newton(func, 0.7, tol=1e-5)





cp = 1
S0 = 100
r = 0.00
tau = 1

K = np.linspace(50,150,10000)
K = np.array(K).reshape([len(K),1])

# COS method settings
L = 3


v0 = 0.0175 # intial variance
vv = 0.5751 # volatility of volatility 
kappa = 1.5768 # rate of mean reversion towards the long-term mean of variance process
theta = 0.0398 # long-term mean of the variance process
rho =-0.5711 # correlation


# The COS method -- Reference value
N = 1000
valCOSRef = heston_cos_vanilla_european(cp, S0, r, tau, K, N, L, v0, vv, kappa, theta, rho)

plt.figure(1)
plt.plot(K,valCOSRef,'-k')
plt.xlabel("strike, K")
plt.ylabel("Option Price")
plt.grid()

# For different numbers of integration terms

N = [32,64,96,128,160,400]
maxError = []
for n in N:
    valCOS = heston_cos_vanilla_european(cp, S0, r, tau, K, n, L, v0, vv, kappa, theta, rho)
    errorMax = np.max(np.abs(valCOS - valCOSRef))
    maxError.append(errorMax)
    print('maximum error for n ={0} is {1}'.format(n,errorMax))
        




# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())

from frm.frm.schedule.daycounter import DayCounter, VALID_DAY_COUNT_BASIS
from frm.frm.schedule.schedule import payment_schedule, VALID_DAY_ROLL, VALID_PAYMENT_TYPE, VALID_PAYMENT_FREQUENCY, VALID_STUB, VALID_STUB_GENERAL, VALID_ROLL_CONVENTION
from frm.frm.schedule.business_day_calendar import get_calendar, VALID_CITY_HOLIDAYS, VALID_CURRENCY_HOLIDAYS
from frm.frm.market_data.iban_ccys import VALID_CCYS

import numpy as np
import pandas as pd
from scipy.optimize import fsolve
from dataclasses import dataclass, field, InitVar 
from typing import Literal, Optional



@dataclass
class CapFloor:
    "Generic class to cover pricing of all types of swap legs and bonds"
    
    # Tenor definitions
    # (i) Specifically define the effective/maturity date of the instrument
    effective_date: Optional[pd.Timestamp]=None
    maturity_date: Optional[pd.Timestamp]=None
    # (ii) Alternatively define the tenor and transcation date/trade date/pricing date 
    #      and the effective_date and maturity_date will be computed
    tenor: Optional[pd.DateOffset]=None, 
    transaction_date: Optional[pd.Timestamp]=None
    forward_starting: int=0 # number of days after the transaction date the instrument starts
    
    transaction_price: Optional[float]=100.0 # As a % of par. 100.0 is the par pricing
    
    # Coupon definitions
    pay_rec: Optional[Literal['pay','rec']]=None
    cap: Optional[float]=None
    floor: Optional[float]=None
    day_count_basis: InitVar[VALID_DAY_COUNT_BASIS] = 'ACT/ACT'
        
    # Notional and currencies definitions
    notional: float=100e6
    
    # Pricing curves
    dsc_crv: Optional['ZeroCurve']=None
    fwd_crv: Optional['ZeroCurve']=None
    
    # Non initialisation arguments
    daycounter: DayCounter = field(init=False)
    
    
    def __post_init__(self, day_count_basis, holiday_calendar):
             
                
        if self.forward_starting is None: self.forward_starting = 0
        if self.roll_convention is None: self.roll_convention='modifiedfollowing'
        if self.stub is None: self.stub = 'first_short'
        if self.payment_type is None: self.payment_type ='in_arrears'
        if self.payment_delay is None: self.payment_delay=0
        if self.fixing_days_ahead is None: self.fixing_days_ahead=2
        if self.notional is None: self.notional=100e6
        if self.exchange_notionals is None: self.exchange_notionals = 'both'
        
        assert bool(self.effective_date is not None and self.maturity_date is not None) \
            != bool(self.tenor is not None and self.transaction_date is not None)
        assert self.fixed_rate is None or self.float_spread is None
        assert self.payment_frequency is not None
        assert self.notional_currency is not None
        assert self.notional > 0
        
        if self.settlement_currency is None: self.settlement_currency = self.notional_currency
        if self.currency_holidays is None: self.currency_holidays = [self.notional_currency, self.settlement_currency]
        if holiday_calendar is None: 
            holiday_calendar = get_calendar(self.currency_holidays, self.city_holidays) 
        if self.effective_date is None: 
            self.effective_date = np.busday_offset(pd.DatetimeIndex([self.transaction_date + pd.DateOffset(days=self.forward_starting)]).values.astype('datetime64[D]'), offsets=0, roll=self.roll_convention, busdaycal=holiday_calendar)[0]
        if self.maturity_date is None: 
            self.maturity_date = np.busday_offset(pd.DatetimeIndex([self.effective_date + self.tenor]).values.astype('datetime64[D]'), offsets=0, roll='following', busdaycal=holiday_calendar)[0]

        self.daycounter = DayCounter(day_count_basis)
                
        if self.fixed_rate is not None: self.leg_type = 'fixed'
        if self.float_spread is not None: self.leg_type = 'float'


# -*- coding: utf-8 -*-

import numpy as np
from scipy.stats import norm
from scipy.optimize import root_scalar, newton


def garman_kohlhagen(S: float, 
                     r_f: float,                    
                     r_d: float,  
                     tau: float, 
                     cp: int = None,
                     σ: float = None,                     
                     K: float = None,
                     F: float = None,                     
                     Δ: float = None,
                     Δ_convention: str = None,
                     task: str='px') -> float:
    """
    GarmanKohlhagen European FX option pricing formula.
    :param S: FX spot, specified in # of units of domestic per 1 unit of foreign currency
    :param tau: time to np.expiry (in years)
    :param r_d: domestic risk free interest rate (annualised)
    :param r_f: foreign riskless interest rate (annualised)
    :param cp: option type (1 for call option (default), -1 for put option)
    :param σ: volatility
    :param K: strike, specified in # of units of term currency per 1 unit of base currency
    :param F: market forward rate, if None is calculated under interest rate parity     
    :param Δ, signed delta, must be in range [0,0.5] for calls, [-0.5,0] for puts
    :param Δ_convention: delta quote convention {'regular_spot_Δ','regular_forward_Δ','premium_adjusted_spot_Δ','premium_adjusted_forward_Δ'}
    :param task: task to perform; 'px' (default), 'strike', 'solve_implied_vol'
    :return: option price (in the domestic currency, per 1 unit of foreign currency notional), 
             strike, calculated from delta
             greeks,
    """
  
    
    S = np.atleast_1d(S).astype(float)
    assert (S > 0.0).all(), S
    
    σ = np.atleast_1d(σ).astype(float)
    # For numerical solving, need to allow negative vol
    # assert (σ > 0.0).all(), σ
    
    tau = np.atleast_1d(tau).astype(float)
    assert (tau > 0.0).all(), tau  
  
    assert task in {'px','strike','greeks'}, task  
  
    if task == 'px':
        assert cp is not None, "'cp' is a required input for task 'px'"
        assert Δ is None, "'Δ' is not an input for task 'px'"
        assert Δ_convention is None, "'Δ_convention' is not an input for task 'px'"
        cp = np.atleast_1d(cp).astype(float)
        assert cp.shape == σ.shape
        assert np.all(np.isin(cp, [1, -1]) | np.isnan(cp))
        d1 = (np.log(S / K) + (r_d - r_f + 0.5 * σ**2) * tau) / (σ * np.sqrt(tau))
        d2 = d1 - σ * np.sqrt(tau)
    elif task == 'strike':
        assert Δ is not None, "'Δ' is a required input for task 'strike'"
        assert Δ_convention is not None, "'Δ_convention' is a required input for task 'strike'"
        assert cp is None, "'cp' is calculated from the 'Δ' input"
        Δ = np.atleast_1d(Δ).astype(float)
        assert Δ.shape == σ.shape
        assert (Δ >= -0.5).all() and (Δ <= 0.5).all()
        cp = np.sign(Δ)
        Δ_convention = Δ_convention.replace('delta','Δ')
        assert Δ_convention in {'regular_spot_Δ',
                                'regular_forward_Δ',
                                'premium_adjusted_spot_Δ',
                                'premium_adjusted_forward_Δ'}, Δ_convention

    if F is not None: 
        F = np.atleast_1d(F).astype(float)
        assert (F > 0.0).all() 
        
    if task == 'px':
        return cp * (S * np.exp(-r_f * tau) * norm.cdf(cp * d1) - K * np.exp(-r_d * tau) * norm.cdf(cp * d2))

    elif task == 'strike':
        result = np.zeros(shape=Δ.shape)
        if F == None:
            F = np.atleast_1d(S * np.exp((r_d - r_f) * tau)) # if not supplied, calculate the forward rate per interest rate parity 
        
        if Δ_convention in {'regular_spot_Δ','regular_forward_Δ'}:
            bool_cond = Δ == 0.5
            if bool_cond.any():
                # at-the-money Δ-neutral strike, for regular spot/forward Δ
                tmp = F * np.exp(0.5 * σ**2 * tau)
                result[bool_cond] = tmp[bool_cond]
            bool_cond = Δ != 0.5
            if bool_cond.any():
                if Δ_convention == 'regular_spot_Δ':
                    tmp = F * np.exp(-cp * norm.ppf(cp * Δ * np.exp(r_f * tau)) * σ * np.sqrt(tau) + (0.5 * σ**2) * tau)
                    result[bool_cond] = tmp[bool_cond]
                elif Δ_convention == 'regular_forward_Δ':
                    tmp = F * np.exp(-cp * norm.ppf(cp * Δ) * σ * np.sqrt(tau) + (0.5 * σ**2) * tau)
                    result[bool_cond] = tmp[bool_cond]
        
        elif Δ_convention in {'premium_adjusted_spot_Δ','premium_adjusted_forward_Δ'}:
            bool_cond = Δ == 0.5
            if bool_cond.any():
                # at-the-money Δ-neutral strike, for premium adjusted spot/forward Δ
                tmp = F * np.exp(-1 * 0.5 * σ**2 * tau)
                result[bool_cond] = tmp[bool_cond]
                
            bool_cond = Δ != 0.5                
            if bool_cond.any(): 
                # For premium adjusted quotes the solution must be solved numerically
                # Please refer to 'A Guide to FX Options Quoting Conventions' by Uwe Wystub for full details
    
                # For the premimum adjusted call Δ, due to non-monotonicity, two strikes can be solved numerically. 
                # To avoid this, we solve bounds to get the correct solution
                
                # 1. Solve the K_max, the upper bound for the numerical solver
                # The premium adjusted Δ is always below the regular (non premium adjusted) Δ
                K_max = garman_kohlhagen(S=S, 
                                         σ=σ, 
                                         r_f=r_f, 
                                         r_d=r_d, 
                                         tau=tau, 
                                         cp=cp, 
                                         Δ=Δ, 
                                         Δ_convention=Δ_convention.replace('premium_adjusted','regular'),
                                         F=F, 
                                         task='strike')  
                
                # 2. Solve K_min, the lower bound for the numerical solver
                # The lower bound is the 'maximum' Δ, hence we numerically solve the maximum Δ
                def solve_K_min(K, σ, cp, F, t):
                    d1 = (np.log(S / K) + (r_d - r_f + 0.5 * σ**2) * t) / (σ * np.sqrt(t))
                    d2 = d1 - σ * np.sqrt(t)
                    return σ * np.sqrt(t) * norm.cdf(d2) - norm.pdf(d2)
                solution = root_scalar(solve_K_min, args=(σ, cp, F, tau), x0=F ,bracket=[0.00001, K_max])
                
                if solution.converged:
                    K_min = solution.root
                else:
                    raise ValueError('the numerical solver for K_min did not converge')
                
                if Δ_convention == 'premium_adjusted_spot_Δ':
                    def solve_Δ(K, σ, cp, F, tau, Δ):
                        return np.exp(r_d * tau) * (cp * K / F) * norm.cdf(cp * (np.log(F/K) - 0.5 * σ ** 2 * tau) / (σ * np.sqrt(tau))) - Δ                
                elif Δ_convention == 'premium_adjusted_forward_Δ':
                    def solve_Δ(K, σ, cp, F, tau, Δ):
                        return (cp * K / F) * norm.cdf(cp * (np.log(F/K) - 0.5 * σ ** 2 * tau) / (σ * np.sqrt(tau))) - Δ
                    
                solution = root_scalar(solve_Δ, args=(σ, cp, F, tau, Δ), x0=F, bracket=[K_min,K_max])
                
                if solution.converged:
                    return solution.root
                else:
                    raise ValueError('the numerical solver for the premium adjusted strike did not converge')
                    
                    
        return result
                    
                
    elif task == 'greeks':
        # INCOMPLETE
        greeks = {}
        
        # Δ appears to be as a % 
        # vega appears to be for 1 unit of term ccy compared to SD, check bbg
        # theta appears to be for 1 unit of base ccy compared to SD, check bbg
        
        # S = 1.1 #EURUSD
        # f = 1.101070 #EURUSD
        # K = 1.101
        # r_f = 0.05 # EUR
        # r_d = 0.1 # USD
        # σ = 0.1
        # days = 7
        # t = days/365
 
        # d1 = (log(S / K) + (r_d - r_f + 0.5 * σ**2) * t) / (σ * np.sqrt(t))
        # d2 = d1 - σ * np.sqrt(t)
    
        
        greeks['spot_Δ'] = cp * np.exp(-r_f * tau) * norm.cdf(cp * d1)
        greeks['fwd_Δ'] = cp * norm.cdf(cp * d1) 
        greeks['vega'] = S * np.sqrt(tau) * norm.pdf(d1) * np.exp(-r_f * tau) # identical for calls and puts
        greeks['theta'] = -(S*np.exp(-r_d*tau)*norm.pdf(d1 * cp)*σ)/(2*np.sqrt(tau)) \
            + r_f * np.exp(-r_f * tau) * S * norm.cdf(d1 * cp) \
            - r_d * np.exp(-r_d * tau) * K * norm.cdf(d2 * cp)
        greeks['gamma'] = np.exp(-r_f * tau) * norm.pdf(d1)  / (S * σ * np.sqrt(tau)) # identical for calls and puts


        return greeks
        
        
def gk_solve_implied_σ(X: float, S: float, K: float, r_d: float, r_f: float, tau: float, cp: int, σ_guess: float) -> float:
    """
    Solve the implied volatility using the Garman-Kohlhagen model for a European Vanilla FX option.

    Parameters:
    - X (float): Option price
    - S (float): Spot price
    - K (float): Strike price
    - r_d (float): Domestic interest rate
    - r_f (float): Foreign interest rate
    - tau (float): Time to maturity in years
    - cp (int): Option type (1 for call, -1 for put)
    - σ_guess (float): Initial guess for volatility

    Returns:
    - float: Implied volatility

    Attempts to find implied volatility using Newton's method initially and falls back to Brent's method in case of failure.
    """
    try:
        # Try Netwon's method first (it's faster but less robust)
        return newton(lambda σ: (garman_kohlhagen(S=S,σ=σ,r_f=r_f,r_d=r_d,tau=tau,cp=cp,K=K,task='px') - X), x0=σ_guess, tol=1e-4, maxiter=50)
    except RuntimeError:
        # Fallback to Brent's method
        try:
            return root_scalar(lambda σ: (garman_kohlhagen(S=S,σ=σ,r_f=r_f,r_d=r_d,tau=tau,cp=cp,K=K,task='px') - X), bracket=[0.0001, 2], method='brentq').root
        except ValueError:
            return np.inf        
        
        
#%%


if __name__ == '__main__':
    # AUD put, put on base ccy
    greeks = garman_kohlhagen(S=0.662866,σ=0.1055564,r_f=0.0466,r_d=0.05381,tau=1.0,cp=-1,K=0.64050,task='greeks')
    px = garman_kohlhagen(S=0.662866,σ=0.1055564,r_f=0.0466,r_d=0.05381,tau=1.0,cp=-1,K=0.64050,task='px')






#%%

# # 2Y USDJPY call, (USD call, JPY Put), 30 June 2023, London 10pm data
# S = 144.32 # 20 Δ Call
# K = 145.83
# σ = 0.0960777
# cp = 1
# r_f = 0.04812
# r_d = -0.00509
# t = 2.0
# F = 129.7958
# Δ = 0.2
# Δ_convention = 'premium_adjusted_fwd_Δ'

# K_IDD = 145.83

# result = garman_kohlhagen(S=S,σ=σ,r_f=r_f,r_d=r_d,t=t,cp=cp,Δ=Δ,F=F,task='strike',Δ_convention=Δ_convention)

# print('result', result)
# print('variance', result - K_IDD)
# print('% variance', 100 * (result - K_IDD) / result)

# #%%

# # 9M USDJPY call, (USD call, JPY Put), 30 June 2023, London 10pm data
# S = 144.32 # 20 Δ Call
# K = 145.83
# σ = 0.0960777
# cp = 1
# r_f = 0.05413
# r_d = -0.00528
# t = 0.75
# F = 138.031
# Δ = 0.2
# Δ_convention = 'premium_adjusted_spot_Δ'

# K_IDD = 148.11

# result = garman_kohlhagen(S=S,σ=σ,r_f=r_f,r_d=r_d,t=t,cp=cp,Δ=Δ,F=F,task='strike',Δ_convention=Δ_convention)

# print('result', result)
# print('variance', result - K_IDD)
# print('% variance', 100 * (result - K_IDD) / result)














